""" pynchon.plugins.Core
"""

from fleks import tagging

from pynchon import abcs, api, cli, models
from pynchon.bin import entry
from pynchon.core import Config as CoreConfig
from pynchon.util import lme

LOGGER = lme.get_logger(__name__)

import fleks

classproperty = fleks.util.typing.classproperty


@tagging.tags(click_aliases=["c"])
class Core(models.Planner):
    """Core Plugin"""

    name = "core"
    priority = -1
    config_class = CoreConfig

    # NB: prevents recursion when `pynchon plan` is used!
    contribute_plan_apply = False

    @classproperty
    def click_group(kls):
        """ """
        kls._finalized_click_groups[kls] = entry
        return kls._finalized_click_groups[kls]

    @classmethod
    def get_current_config(kls):
        """ """
        from pynchon import config as config_mod

        result = getattr(config_mod, kls.get_config_key())
        return result

    def cfg(self):
        """Show current project config (with templating/interpolation)"""
        tmp = self.project_config
        return tmp

    @cli.click.flag("--bash", help="bootstrap bash")
    @cli.click.flag("--bashrc", help="bootstrap bashrc")
    @cli.click.flag("--bash-completions", help="bootstrap completions")
    @cli.click.flag("--pynchon", help="bootstrap .pynchon.json5")
    @cli.click.flag("--makefile", help="bootstrap Makefile")
    @cli.click.flag("--tox", help="bootstrap tox")
    def bootstrap(
        self,
        pynchon: bool = False,
        bash: bool = False,
        bash_completions: bool = False,
        bashrc: bool = False,  # noqa
        makefile: bool = False,
        tox: bool = False,
    ) -> None:
        """Bootstrap for shell integration, etc"""
        template_prefix = f"{self.plugin_templates_prefix}/bootstrap"
        pynchon_completions_script = ".tmp.pynchon.completions.sh"
        bashrc_snippet = ".tmp.pynchon.bashrc"
        if bash_completions:
            pass

            gr = self.__class__.click_entry
            all_known_subcommands = [
                " ".join(x.split()[1:])
                for x in cli.click.subcommand_tree(
                    gr, mode="text", path=tuple(["pynchon"]), hidden=False
                ).keys()
            ]
            # head = [x for x in all_known_subcommands if len(x.split()) == 1]
            # rest = [x for x in all_known_subcommands if x not in head]
            # tmp = collections.defaultdict(list)
            # for phrase in rest:
            #     bits = phrase.split()
            #     k = bits.pop(0)
            #     tmp[k] += bits
            # print(list(tmp.keys()))
            print(all_known_subcommands)
            return
        #   if bash:
        #       rest = [
        #           f"""    '{k}'*)
        #         while read -r; do COMPREPLY+=( "$REPLY" ); done < <( compgen -W "$(_pynchon_completions_filter "{' '.join(subs)}")" -- "$cur" )
        #         ;;
        #       """
        #           for k, subs in tmp.items()
        #       ]
        #       rest += [
        #           f"""    *)
        # while read -r; do COMPREPLY+=( "$REPLY" ); done < <( compgen -W "$(_pynchon_completions_filter "{' '.join(head)}")" -- "$cur" )
        # ;;"""
        #       ]
        #       # LOGGER.warning("This is intended to be run through a pipe, as in:")
        #       # LOGGER.critical("pynchon bootstrap --bash | bash")
        #       tmpl = api.render.get_template(f"{template_prefix}/bash.sh")
        #       content = tmpl.render(head=head, rest="\n".join(rest), **self.config)
        #       files.dumps(content=content, file=pynchon_completions_script)
        #       LOGGER.warning(
        #           f"To refresh your shell, run: `source {pynchon_completions_script}`"
        #       )
        #       return dict()
        # if bashrc:
        #     LOGGER.critical(
        #         "To use completion hints every time they are "
        #         "present in a folder, adding this to .bashrc:"
        #     )
        #     tmpl = api.render.get_template(f"{template_prefix}/bashrc.sh")
        #     content = tmpl.render(pynchon_completions_script=pynchon_completions_script)
        #     files.dumps(content=content, file=bashrc_snippet, logger=LOGGER.info)
        #     return files.block_in_file(
        #         target_file=abcs.Path("~/.bashrc").expanduser(),
        #         block_file=bashrc_snippet,
        #     )
        elif pynchon:
            tmp = abcs.Path(".") / ".pynchon.json5"
            if tmp.exists():
                err = f"Cowardly refusing to recreate {tmp}"
                LOGGER.critical(err)
                raise SystemExit(1)
            else:
                print("pynchon pattern sync . docs --plan")
        elif bash:
            this_cmd = "pynchon bootstrap --bash"  # FIXME: get from click-ctx
            LOGGER.debug("collecting `shell_aliases` from all plugins")
            out = self.siblings.collect_config_dict("shell_aliases")
            out = "\n".join([f"alias {k}='{v}';" for k, v in out.items()])
            print(out)
            LOGGER.warning(f'for this session, use "source <({this_cmd})"')
            LOGGER.warning(f'for it to be permanent, use "{this_cmd} >> ~/.bashrc"')
        elif tox or makefile:
            tail = "Makefile" if makefile else "tox.ini"
            tmpl = api.render.get_template(f"{template_prefix}/{tail}")
            content = tmpl.render(**self.project_config.dict())
            print(content)

    def raw(self) -> None:
        """
        Returns (almost) raw config, without interpolation
        """
        from pynchon.config import RAW

        print(RAW.json())

    def plan(
        self,
        config=None,
    ) -> models.Plan:
        """Runs plan for all plugins

        :param config: Default value = None)

        """

        config = config or self.project_config
        plan = super(self.__class__, self).plan(config)
        plugins = self.siblings.values()
        plugins = [
            p
            for p in plugins
            if isinstance(p, models.AbstractPlanner) and p.contribute_plan_apply
        ]
        self.logger.critical(f"Planning on behalf of: {[p.name for p in plugins]}")
        for plugin_obj in plugins:
            subplan = plugin_obj.plan()
            if not subplan:
                self.logger.warning(
                    f"subplan for '{plugin_obj.__class__.__name__}' is empty!"
                )
            else:
                for g in subplan.goals:
                    self.logger.info(f"{plugin_obj} contributes {g}")
                    plan.append(g)
        return plan.finalize()



""" pynchon.plugins.python.api
"""

from fleks import cli, tagging

from pynchon import abcs, models
from pynchon.api import render
from pynchon.util import complexity, lme, typing

LOGGER = lme.get_logger(__name__)


@tagging.tags(click_aliases=["pa"])
class PythonAPI(models.Planner):
    """Generators for Python API docs"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "python-api"
        skip_private_methods: bool = typing.Field(default=True)
        skip_patterns: typing.List[str] = typing.Field(default=[])
        apply_hooks: typing.List[str] = typing.Field(default=["open-after"])

    name = "python-api"

    @cli.click.group("gen")
    def gen(self):
        """Generates API docs from python modules, packages, etc"""

    # FIXME: not bound correctly: missing 1 req pos arg 'self'
    @gen.command("toc")
    @cli.options.file
    @cli.options.header
    @cli.options.should_print
    @cli.options.package
    @cli.click.option(
        "--output",
        "-o",
        default="docs/api/README.md",
        help=("output file to write.  (optional)"),
    )
    @cli.click.option(
        "--exclude",
        default="",
        help=("comma-separated list of modules to exclude (optional)"),
    )
    def toc(
        self,
        package=None,
        should_print=None,
        file=None,
        exclude=None,
        output=None,
        stdout=None,
        header=None,
    ):
        """Generate table-of-contents"""

        T_TOC_API = render.get_template("pynchon/plugins/python/api/TOC.md.j2")
        module = complexity.get_module(package=package, file=file)
        result = complexity.visit_module(
            module=module,
            module_name=package,
            template=T_TOC_API,
            exclude=exclude.split(","),
        )
        header = f"{header}\n\n" if header else ""
        result = dict(
            header=f"## API for '{package}' package\n\n{header}" + "-" * 80,
            blocks=result,
        )
        result = result["header"] + "\n".join(result["blocks"])
        print(result, file=open(output, "w"))
        if should_print and output != "/dev/stdout":
            print(result)

    def plan(self, config=None) -> typing.List:
        """
        Runs a plan for this plugin
        """
        config = config or self.project_config
        plan = super(self.__class__, self).plan(config)
        docs_root = self[:"docs.root":]
        api_docs_root = f"{docs_root}/api"
        if not abcs.Path(api_docs_root).exists():
            plan.append(
                self.goal(
                    command=f"mkdir -p {api_docs_root}",
                    resource=api_docs_root,
                    type="mkdir",
                )
            )
        pkg = self[:"python.package.name":None]
        pkg_arg = pkg and f"--package {pkg}"
        src_root = self[:"src.root":]
        src_arg = src_root and f"--file {src_root}"
        input = f"{pkg_arg or src_arg}"
        outputf = f"{api_docs_root}/README.md"
        output = f"--output {outputf}"
        cmd_t = "pynchon python-api gen toc"
        plan.append(
            self.goal(
                resource=outputf,
                command=f"{cmd_t} {input} {output}",
                type="gen",
                # ordering=f'{len(plan)+1}/?',
            )
        )
        return plan.finalize()



""" pynchon.plugins.src
"""

import fnmatch

from pynchon import abcs, api, cli, events, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)

EXT_MAP = {
    ".ini": dict(
        template="pynchon/plugins/src/header/ini.j2", pre=["#", "###"], post="###"
    ),
    ".j2": dict(
        template="pynchon/plugins/src/header/jinja.j2", pre=["{#", "<!--"], post="#}"
    ),
    "*.md.j2": dict(
        template="pynchon/plugins/src/header/jinja-md.md.j2",
        pre=["{#", "<!--"],
        post="#}",
    ),
    ".json5": dict(
        template="includes/pynchon/src/json5-header.j2", pre=["//", "///"], post="///"
    ),
    ".py": dict(
        # template='includes/pynchon/src/python-header.j2',
        template="pynchon/plugins/src/header/python.j2",
        pre=['"""', '"', "'"],
        post='""""',
    ),
    ".sh": dict(
        template="pynchon/plugins/src/header/sh.j2", pre=["#", "###"], post="###"
    ),
}


class SourceMan(models.ResourceManager):
    """Management tool for project source"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "src"
        goals: typing.List[str] = typing.Field(default=[])
        include_patterns: typing.List[str] = typing.Field(default=[])
        exclude_patterns: typing.List[str] = typing.Field(default=[])
        root: typing.Union[str, abcs.Path, None] = typing.Field(default=None)
        sorted: bool = typing.Field(default=False, help="Whether to sort source code")

    name = "src"
    cli_name = "src"
    priority = 0

    # @tagging.tagged_property(conflict_strategy="override")
    # @property
    # def exclude_patterns(self):
    #     from pynchon.plugins import util as plugin_util
    #
    #     globals = plugin_util.get_plugin("globals").get_current_config()
    #     global_ex = globals["exclude_patterns"]
    #     my_ex = self.get("exclude_patterns", [])
    #     return list(set(global_ex + my_ex + ["**/pynchon/templates/includes/**"]))

    # @tagging.tagged_property(conflict_strategy='override')
    # def exclude_patterns(self):
    #     globals = plugin_util.get_plugin('globals').get_current_config()
    #     global_ex = globals['exclude_patterns']
    #     my_ex = self.get('exclude_patterns', [])
    #     return list(set( global_ex+my_ex))

    # def list(self):
    #     """ """
    #     # config = api.project.get_config()
    #     # src_root = config.pynchon['src_root']
    #     include_patterns = self.config.get('include_patterns', ["**"])
    #     return files.find_globs(include_patterns)
    def list_modified(self):
        """
        Lists modified files
        """
        raise NotImplementedError()

    def _get_meta(self, rsrc):
        tmp = rsrc.full_extension()
        try:
            ext_meta = EXT_MAP[tmp]
        except (KeyError,) as exc:
            for x in EXT_MAP:
                if fnmatch.fnmatch(tmp, x):
                    ext_meta = EXT_MAP[x]
                    break
            else:
                LOGGER.warning(f"no match for {x}")
                return
        return ext_meta

    def _get_missing_headers(self, resources):
        """
        :param resources:
        """
        result = dict(extensions=set(), files=[])
        for p_rsrc in resources:
            if not p_rsrc.is_file() or not p_rsrc.exists():
                continue
            # ext_info = self._rsrc_ext_info(p_rsrc)
            tmp = p_rsrc.full_extension()
            ext_meta = self._get_meta(p_rsrc)
            if ext_meta is None:
                continue
            preamble_patterns = ext_meta["pre"]
            assert isinstance(preamble_patterns, (list,))
            with p_rsrc.open("r") as fhandle:
                content = fhandle.read().lstrip()
                if any([content.startswith(pre.lstrip()) for pre in preamble_patterns]):
                    # we detected expected comment at the top of the file,
                    # so the appropriate header *might* be present; skip it
                    continue
                else:
                    # no header at all
                    result["files"].append(p_rsrc)
                    result["extensions"] = result["extensions"].union(
                        {p_rsrc.full_extension()}
                    )
        result.update(extensions=list(result["extensions"]))
        return result

    def _plan_empties(self, resources):
        """
        :param resources:
        """
        result = []
        return result

    def _render_header_file(self, rsrc: abcs.Path = None):
        """
        :param rsrc: abcs.Path:  (Default value = None)
        """
        ext = rsrc.full_extension()
        templatef = self._get_meta(rsrc)["template"]
        tpl = api.render.get_template(templatef)
        abs = rsrc.absolute()
        src_root = abcs.Path(self.config["root"]).absolute()
        try:
            relf = abs.relative_to(src_root)
        except ValueError:
            relf = abs.relative_to(abcs.Path(".").absolute())
        relf = relf.path_truncated()
        module_dotpath = str(relf).replace("/", ".")
        tmp2 = __name__.replace(".", "-")
        fname = f".tmp.src-header.{module_dotpath}{ext}"
        result = tpl.render(
            module_dotpath=module_dotpath,
            template=templatef,
            filename=str(abs),
            relative_filename=relf,
        )
        if not result:
            err = f'header for extension "{ext}" rendered to "{fname}" from {templatef}'
            self.logger.warning(err)
        with open(fname, "w") as fhandle:
            fhandle.write(result)
        LOGGER.warning(f"wrote {fname}")
        return fname

    @cli.click.group("open")
    def _open(self):
        """helper for opening project source files"""

    @_open.command("recent")
    def open_recent(self):
        """opens recently changed files"""

    @_open.command("changed")
    def open_changed(self):
        """opens changed files"""

    def plan(self, config=None):
        """Describe plan for this plugin"""
        plan = super().plan(config=config)
        resources = [abcs.Path(fsrc) for fsrc in self.list()]
        self.logger.warning("Adding user-provided goals")
        for g in self["goals"]:
            plan.append(self.goal(command=g, resource="?", type="user-config"))

        self.logger.warning("Adding file-header related goals")
        cmd_t = "python -mpynchon.util.files prepend --clean "
        loop = self._get_missing_headers(resources)
        for rsrc in loop["files"]:
            if rsrc.match_any_glob(self["exclude_patterns"::[]]):
                continue
            ext = rsrc.full_extension()
            ext = ext[1:] if ext.startswith(".") else ext
            # fhdr = header_files[ext]
            fhdr = self._render_header_file(rsrc)
            plan.append(
                self.goal(
                    resource=rsrc,
                    type="change",
                    label=f"Adding file header for '{ext}'",
                    command=f"{cmd_t} {fhdr} {rsrc}",
                )
            )
        # for rsrc in self._plan_empties(resources):
        #     plan.append(
        #         self.goal(
        #             resource=rsrc,
        #             type='delete',
        #             command=f'rm {rsrc}',
        #         )
        #     )
        return plan

    def find(self):
        """file finder"""

    def header(self):
        """creates file headers for source in {src_root}"""

    def map(self):
        """file mapper"""



""" pynchon.config
"""

# from pynchon import abcs
from fleks.plugin import Meta

from pynchon.api import render
from pynchon.app import app
from pynchon.core import Config as CoreConfig
from pynchon.util import lme  # typing

from pynchon.plugins.git import GitConfig  # noqa

from .util import config_folders  # noqa
from .util import finalize  # noqa
from .util import get_config_files  # noqa
from .util import load_config_from_files  # noqa

LOGGER = lme.get_logger(__name__)
events = app.events

# FIXME: abstract into phases inside pynchon.app
msg = "Loading raw-config from OS.."
events.lifecycle.send(__name__, stage=msg)
git = GIT = GitConfig()
# import IPython; IPython.embed()
msg = "Building raw-config from files.."
events.lifecycle.send(
    __name__,
    msg=msg,
    stage=msg,
)
CONFIG_FILES = []
MERGED_CONFIG_FILES = {}
for cfg_src, config in load_config_from_files().items():
    MERGED_CONFIG_FILES = {**MERGED_CONFIG_FILES, **config}
    config and CONFIG_FILES.append(cfg_src)

# NB: this content is potentially templated
msg = "Building plugins-list.."
events.lifecycle.send(
    __name__,
    msg=msg,
    stage=msg,
)

pynchon = PYNCHON = CoreConfig(
    skip_instance_validation=True,  # still bootstrapping..
    config_files=CONFIG_FILES,
    **MERGED_CONFIG_FILES,
)
RAW = PYNCHON.copy()
PLUGINS = PYNCHON["plugins"]
LOGGER.warning(f"plugins: {PLUGINS}")
# FIXME: get from registry or mcls
_all_names = PLUGINS + Meta.NAMES

msg = "Splitting core config.."
events.lifecycle.send(
    __name__,
    msg=msg,
    stage=msg,
)
PYNCHON_CORE = {x: PYNCHON[x] for x in PYNCHON.__fields__.keys() if x not in _all_names}
PYNCHON_CORE = CoreConfig(**PYNCHON_CORE)

msg = "Interpolating config.."
events.lifecycle.send(
    __name__,
    msg=msg,
    stage=msg,
)

USER_DEFAULTS = render.dictionary(input=RAW.copy(), context=dict(pynchon=RAW))



""" pynchon.plugins.scaffolding
"""

import os

from pynchon import abcs
from pynchon.util import lme, typing

LOGGER = lme.get_logger(__name__)


class ScaffoldingItem(abcs.AttrDict):
    warnings = []

    def __init__(
        self, name="unnamed scaffold", scope="*", pattern=None, src=None, **kwargs
    ):
        assert pattern is not None
        assert src is not None
        super().__init__(
            name=name,
            scope=scope,
            src=abcs.Path(os.path.expanduser(src)),
            pattern=pattern,
            **kwargs,
        )
        self.validate()

    @property
    def exists(self) -> typing.Bool:
        return os.path.exists(self.src)

    def validate(self):
        if not self.exists and self.src not in ScaffoldingItem.warnings:
            LOGGER.critical(f"Scaffolding source @ {self.src} does not exist!")
            ScaffoldingItem.warnings.append(self.src)


class ScaffoldingConfig(abcs.Config):
    """ """

    config_key: typing.ClassVar[str] = "scaffolding"

    # def scaffolds(self):
    #     return dict([
    #         [scaffold_pattern,scaffold_meta]
    #         for k,v in self.items()])



""" pynchon.plugins.git
"""

from fleks import tagging

from pynchon import abcs, models
from pynchon.util import files, lme, os, typing

LOGGER = lme.get_logger(__name__)


class GitConfig(abcs.Config):
    """ """

    # _root: str = None
    config_key: typing.ClassVar[str] = "git"

    # class Config:
    #     fields = {
    #         # '_root': 'root',
    #         '_repo': 'repo',
    #     }

    def _run(self, cmd, log_command=False, **kwargs):
        """
        :param cmd: param log_command:  (Default value = False)
        :param log_command:  (Default value = False)
        :param **kwargs:
        """
        if self.root:
            pre = f"cd {self.root} && " if self.root else ""
            return os.invoke(f"{pre}{cmd}", log_command=log_command, **kwargs)

    #
    # @memoized_property
    # def default_remote_branch(self) -> typing.StringMaybe:
    #     """ """
    #     tmp = self._run("git remote show origin " "| sed -n '/HEAD branch/s/.*: //p'")
    #     if tmp and tmp.succeeded:
    #         return tmp.stdout.strip() or None

    @property
    def root(self) -> typing.StringMaybe:
        """ """
        tmp = self.__dict__.get("_root")
        if tmp:
            return tmp
        tmp = files.get_git_root(abcs.Path("."))
        return tmp and tmp.parents[0]

    @property
    def repo(self) -> typing.StringMaybe:
        """ """
        if "repo" not in self.__dict__:
            cmd = self._run("git config --get remote.origin.url")
            self.__dict__.update(
                repo=cmd and (cmd.stdout.strip() if cmd.succeeded else None)
            )
        return self.__dict__["repo"]

    @property
    def is_github(self):
        """ """
        tmp = "git@github https://github.com http://github.com".split()
        return self.repo and any([self.repo.startswith(x) for x in tmp])

    @property
    def github_org(self) -> typing.StringMaybe:
        """ """
        if self.is_github:
            tmp = self.repo.split(":")[-1]
            try:
                org, _repo_name = tmp.split("/")
            except (ValueError,):
                return None
            return org

    @property
    def repo_name(self) -> typing.StringMaybe:
        """ """
        if self.repo:
            tmp = self.repo.split(":")[-1]
            try:
                _org, repo_name = tmp.split("/")
            except (ValueError,):
                return None
            repo_name = repo_name.split(".git")[0]
            return repo_name

    @property
    def repo_url(self):
        """ """
        if all([self.github_org, self.repo_name]):
            return f"https://github.com/{self.github_org}/{self.repo_name}"

    @property
    def branch_name(self):
        """ """
        if "branch_name" not in self.__dict__:
            cmd = self._run("git rev-parse --abbrev-ref HEAD")
            tmp = cmd and cmd.succeeded and cmd.stdout.strip()
            self.__dict__.update(branch_name=tmp or None)
        return self.__dict__["branch_name"]

    @property
    def hash(self) -> str:
        """ """
        if "hash" not in self.__dict__:
            cmd = self._run("git rev-parse HEAD")
            tmp = cmd and cmd.succeeded and cmd.stdout.strip()
            self.__dict__.update(hash=tmp or None)
        return self.__dict__["hash"]


@tagging.tags(click_aliases=["g"])
class Git(models.Provider):
    """Context for git"""

    priority = -2
    name = "git"
    config_class = GitConfig

    @tagging.tags(click_aliases=["ls"])
    def list(self, changes=False) -> typing.List[abcs.Path]:
        """lists files tracked by git"""
        if changes:
            return self.status()["modified"]
        else:
            cmd = self.config._run("git ls-files")
            lines = [line.lstrip().strip() for line in cmd.stdout.split("\n")]
            lines = [filter(None, line.split(" ")) for line in lines if line]
            return [abcs.Path(p) for p in lines]

    @property
    def modified(self) -> typing.List[abcs.Path]:
        """ """
        return self.status().get("modified", [])

    @tagging.tags(click_aliases=["st", "stat"])
    def status(self) -> typing.Dict:
        """JSON version of `git status` for this project"""
        cmd = self.config._run("git status --short")
        lines = [line.lstrip().strip() for line in cmd.stdout.split("\n")]
        lines = [list(filter(None, line.split(" "))) for line in lines if line]
        lines = [line for line in lines if len(line) == 2]
        abspaths = []
        for code, fname in lines:
            abspaths.append((code, abcs.Path(self.config.root) / abcs.Path(fname)))
        modified = [p for (code, p) in abspaths if code.strip() == "M"]
        return dict(modified=modified)



""" pynchon.plugins.docs.main
"""

import fleks
import gripe
from fleks import cli, tagging
from memoized_property import memoized_property

from pynchon.util.os import invoke
from pynchon.plugins.docs.opener import OpenerMixin

from pynchon import abcs, api, events, models  # noqa
from pynchon.util import files, lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


@tagging.tags(click_aliases=["d"])
class DocsMan(models.ResourceManager, OpenerMixin):
    """
    Management tool for project docs, including helpers for enumerating, serving, & opening them.
    """

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "docs"
        include_patterns: typing.List[str] = typing.Field(default=[])
        root: typing.Union[str, abcs.Path, None] = typing.Field()
        exclude_patterns: typing.List[str] = typing.Field(default=[])

    name = "docs"
    cli_name = "docs"
    cli_label = "Manager"
    priority = 0
    serving = None

    # @property
    # def root(self):
    #     if "root" not in self.__dict__:
    #         from pynchon.config import GIT, pynchon
    #         tmp = GIT.root
    #         self.__dict__.update(root=tmp or pynchon["working_dir"])
    #     return self.__dict__["root"]

    @property
    def server_pid(self):
        tmp = self.server.proc
        return tmp and tmp.pid

    @property
    def server_url(self):
        return f"http://localhost:{self.server.port}"

    @property
    def git_root(self):
        return self[: "git.root" : self[:"pynchon.working_dir"]]

    @memoized_property
    def server(self):
        return gripe.server

    @cli.click.group("gen")
    def gen(self):
        """Generator subcommands"""

    # @cli.click.option('--output',default=None)
    @cli.options.should_print
    @gen.command("version-file")
    def version_file(
        self,
        should_print: bool,
        # output:str=None,
    ):
        """
        Creates {docs.root}/VERSION.md file
        """
        output = abcs.Path(self[:"docs.root":]) / "VERSION.md"
        tmpl = api.render.get_template("pynchon/plugins/core/VERSIONS.md.j2")
        result = tmpl.render(**self.project_config.dict())
        files.dumps(
            content=result, file=output, quiet=False, logger=self.logger.warning
        )
        if should_print and output != "/dev/stdout":
            print(result)
        return True

    @cli.click.option("--background", is_flag=True, default=True)
    @cli.click.option("--force", is_flag=True, default=False)
    def serve(
        self,
        # background: bool = True,
        force: bool = False,
    ) -> typing.Dict[str, str]:
        """
        Runs a `gripe` server for this project
        """
        LOGGER.critical("running serve")
        args = "--force" if force else ""
        if not self.server.live or force:
            cmd = f"python -m gripe start {args}"
            LOGGER.critical(cmd)
            assert invoke(cmd).succeeded
        return dict(url=self.server_url, pid=self.server_pid)

    @tagging.tags(click_aliases=["op", "opn"])
    @fleks.cli.arguments.file
    def open(self, file, server=None):
        """Opens a docs-artifact (based on file type)"""
        self.serve()
        file = abcs.Path(file)
        if not file.exists():
            raise ValueError(f"File @ `{file}` does not exist")
        ext = file.full_extension().replace(".", "_")
        opener = f"_open_{ext}"
        try:
            fxn = getattr(self, opener)
        except (AttributeError,) as exc:
            err = f"dont know how to open `{file}`, " f"method `{opener}` is missing"
            LOGGER.warning(err)
            ext = file.full_extension().split(".")[-1]
            opener = f"_open__{ext}"
            try:
                fxn = getattr(self, opener)
            except (AttributeError,) as exc:
                err = (
                    f"dont know how to open `{file}`, " f"method `{opener}` is missing"
                )
                raise NotImplementedError(err)
        return fxn(file)

    def open_changes(self) -> typing.List[str]:
        """Opens changed files"""
        result = []
        changes = self.list(changes=True)
        if not changes:
            LOGGER.warning("No changes to open")
            return []
        LOGGER.warning(f"opening {len(changes)} changed files..")
        if len(changes) > 10:
            LOGGER.critical("10+ changes; refusing to open this many.")
            return result
        for ch in changes:
            LOGGER.warning(f"opening {ch}")
            result.append(self.open(ch))
        return result

    def plan(self, config=None):
        """Creates a plan for this plugin"""
        plan = super(self.__class__, self).plan(config=config)
        rsrc = self.config["root"]
        if not abcs.Path(rsrc).exists():
            plan.append(
                self.goal(resource=rsrc, type="mkdir", command=f"mkdir -p {rsrc}")
            )
        cmd_t = f"{self.cli_path} gen version-file"
        rsrc = abcs.Path(rsrc) / "VERSION.md"
        plan.append(self.goal(resource=rsrc, type="gen", command=f"{cmd_t}"))
        return plan



""" pynchon.abcs.attrdict
"""

from pynchon.util import lme, typing

LOGGER = lme.get_logger(__name__)


class AttrDictBase:
    """A dictionary with attribute-style access.
    It maps attribute access to the real dictionary.


    """

    def __init__(self, **init: typing.OptionalAny):
        dict.__init__(self, init)

    def __getattr__(self, name: str) -> typing.Any:
        try:
            return super().__getitem__(name)
        except (KeyError,) as exc:
            LOGGER.info(f"AttrDict: KeyError accessing {name}")
            raise AttributeError(exc)

    def __setitem__(self, key: str, value: typing.Any) -> typing.Any:
        return super().__setitem__(key, value)

    __setattr__ = __setitem__

    def __getitem__(self, name: str) -> typing.Any:
        try:
            return super().__getitem__(name)
        except (KeyError,) as exc:
            LOGGER.info(f"AttrDict: KeyError accessing {name}")
            raise KeyError(exc)

    def __delitem__(self, name: str) -> typing.Any:
        return super().__delitem__(name)

    def __getstate__(self) -> typing.Iterable:
        return self.__dict__.items()

    def __setstate__(self, items: typing.Any) -> None:
        for key, val in items:
            self.__dict__[key] = val

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({dict.__repr__(self)})"

    # def as_dict(self):
    #     return dict([[key,getattr(self,key)] for key in self])


class AttrDict(AttrDictBase, dict):
    pass



""" pynchon.util.text.dumpf.__main__
"""

from fleks.cli import click, options  # noqa
from fleks.util.tagging import tags

from pynchon.cli import common

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)

entry = common.entry_for(__name__)


import shimport

tmp = (
    shimport.wrapper("pynchon.util.text.dumpf")
    .prune(
        filter_instances=typing.FunctionType,
        filter_module_origin="pynchon.util.text.dumpf",
    )
    .map_ns(
        lambda _name, fxn: [fxn, tags[fxn].get("click_aliases", []) + [fxn.__name__]]
    )
    .starmap(
        lambda fxn, aliases: [
            common.kommand(
                name=alias.replace("_", "-"),
                parent=entry,
                help=(
                    fxn.__doc__
                    if alias == fxn.__name__
                    else f"alias for `{fxn.__name__}`"
                ),
            )(fxn)
            for alias in aliases
        ],
        logger=LOGGER,
    )
)
LOGGER.debug(tmp)

if __name__ == "__main__":
    entry()



""" pynchon.plugins
"""

import collections

import fleks
import shimport

from pynchon import config, events  # noqa
from pynchon.util import lme, typing  # noqa

from .util import get_plugin, get_plugin_obj  # noqa

LOGGER = lme.get_logger(__name__)
mod_wrap = shimport.wrap(__name__, import_children="**/*.py")
registry = {
    **mod_wrap.prune(
        exclude_names="git".split(),  # FIXME: hack
        types_in=[fleks.Plugin],
        filter_vals=[
            lambda val: val.name in config.PLUGINS,
        ],
    ).namespace
}
registry = registry.items()
registry = collections.OrderedDict(sorted(registry, key=lambda x: x[1].priority))
registry = collections.OrderedDict(
    [
        [plugin_kls.name, dict(obj=None, kls=plugin_kls)]
        for k, plugin_kls in registry.items()
    ]
)
registry["core"]["obj"] = registry["core"]["kls"]()
# registry = collections.OrderedDict(sorted(registry.items(), key=lambda x: x[1]['kls'].priority))
events.lifecycle.send(__name__, msg="Finished creating plugin registry")



""" pynchon.abcs.visitor
"""

import copy
from types import MappingProxyType

import pydash

from pynchon import abcs
from pynchon.util import lme

LOGGER = lme.get_logger(__name__)


class Visitor:
    def __init__(
        self,
        filter_path=lambda _: True,
        filter_value=lambda _: True,
        trigger=lambda p, v: (p, v),
        paths=[],
        obj=None,
        **kwargs,
    ):
        self.conf = kwargs
        self.filter_path = filter_path
        self.filter_value = filter_value
        self.trigger = trigger

    def matched(self, path=None, value=None):
        """
        :param path: Default value = None)
        :param value: Default value = None)
        """
        default = self.trigger(path, value)
        if "accumulate_paths" in self.conf:
            return path
        elif "accumulate_values" in self.conf:
            return value
        else:
            return default

    def __call__(
        self,
        path=None,
        value=None,
    ):
        """
        :param path: Default value = None)
        :param value: Default value = None)
        """
        if all([self.filter_path(path), self.filter_value(value)]):
            return self.matched(path=path, value=value)


def traverse(obj, visitor=None, visitor_kls=None, visitor_kwargs={}):
    """# example `visitor`:
    #
    # def visit(value=None, path=None):
    #     LOGGER.debug(f"[{path}: {value}]")
    #     return value

    :param obj: param visitor:  (Default value = None)
    :param visitor_kls: Default value = None)
    :param visitor_kwargs: Default value = {})
    :param visitor:  (Default value = None)

    """
    assert bool(visitor) ^ bool(visitor_kls) ^ bool(visitor_kwargs)
    paths = []

    def travel(arg=obj, path="", paths=[]):
        if isinstance(arg, (list, tuple)):
            for i, item in enumerate(arg):
                path_key = f"{path}.{i}"
                paths += [path_key]
                travel(item, path=path_key)
        if isinstance(arg, (dict,)):
            for k, v in arg.items():
                path_key = f"{path}.{k}"
                paths += [path_key]
                travel(v, path=path_key)
        paths = sorted(list(set(paths)))
        return paths

    paths = travel()
    result = dict(paths=paths, obj=obj)
    if visitor_kwargs and not visitor_kls:
        visitor_kls = Visitor
    if visitor_kls:
        assert not visitor
        visitor = visitor_kls(**{**result, **visitor_kwargs})
    visits = []
    if visitor is not None:
        for path in paths:
            visitation = visitor(path=path, value=pydash.get(obj, path))
            visitation and visits.append(visitation)
    result.update(
        # visitor=visitor,
        visits=visits
    )
    result = abcs.AttrDict(**result)
    return result


class TemplatedDict(dict):
    """ """

    def __init__(self, dct):
        """

        :param dct:

        """
        super().__init__(dct.copy())
        self.logger = lme.get_logger(self.__class__.__name__)

    def get_path(self, path):
        """

        :param path:

        """
        return pydash.get(self, path)

    def set_path(self, path, val):
        """

        :param path: param val:
        :param val:

        """
        return pydash.set_with(self, path, val)

    @property
    def traversal(self):
        """ """
        traversed = traverse(
            self,
            visitor_kwargs=dict(filter_value=self.is_templated, accumulate_paths=True),
        )
        return traversed

    @property
    def unresolved(self):
        """ """
        return self.traversal.visits


import jinja2

# from pynchon.api.render import UndefinedError


class JinjaDict(TemplatedDict):
    """ """

    def render(self, ctx={}):
        """
        :param ctx: Default value = {})
        """
        tmp = copy.deepcopy(self)
        while tmp.unresolved:
            templated = tmp.unresolved
            self.logger.debug(f"remaining unresolved: {len(templated)}")
            for i, path in enumerate(templated):
                templated.pop(i)
                val = self.get_path(path)
                try:
                    x = tmp.render_path(path, ctx=ctx)
                    rspec = dict(resolution=x, path=path, value=val)
                    LOGGER.info(rspec)
                except (jinja2.exceptions.UndefinedError,) as exc:
                    self.logger.debug(f"resolution for {path}@`{val}` failed ({exc})")
                    self.logger.debug(exc)
                    # self.logger.debug(f"self: {tmp}")
                    # self.logger.debug(f"ctx: {ctx}")
                    templated.append(path)
                else:
                    break
            else:
                break
        return MappingProxyType(tmp)

    def render_path(self, path, ctx={}, strict=False):
        """

        :param path: param ctx:  (Default value = {})
        :param strict: Default value = False)
        :param ctx:  (Default value = {})

        """
        # from pynchon.api import render
        from pynchon.util.text import render

        strict and True
        value = self.get_path(path)
        resolved = render.jinja(
            text=value,
            context={**self, **ctx},
            # includes=[],
        )
        self.set_path(path, resolved)
        return resolved

    def is_templated(self, v):
        """

        :param v:

        """
        return isinstance(v, (str,)) and "{{" in v



""" pynchon.plugins.__template__
"""

from pynchon import abcs, cli, events, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class PluginTemplate(models.Provider):
    """PluginTemplate"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "template"

    name = "template"
    cli_name = "template"
    cli_label = "Meta"



""" {{pkg}}.util.classproperty
"""

from . import typing


@typing.validate_arguments
def new_in_class(name: str, kls: typing.Type):
    """

    :param name: str:
    :param kls: typing.Type:
    :param name: str:
    :param kls: typing.Type:

    """
    return name in dir(kls) and not any([name in dir(base) for base in kls.__bases__])


def is_subclass(x, y, strict=True):
    """

    :param x: param y:
    :param strict: Default value = True)
    :param y:

    """
    if isinstance(x, (typing.Type)) and issubclass(x, y):
        if strict and x == y:
            return False
        return True
    return False


class classproperty:
    """ """

    def __init__(self, fxn):
        self.fxn = fxn

    def __get__(self, obj, owner) -> typing.OptionalAny:
        return self.fxn(owner)


class classproperty_cached(classproperty):
    """ """

    CLASSPROP_CACHES = {}

    def __get__(self, obj, owner) -> typing.OptionalAny:
        result = self.__class__.CLASSPROP_CACHES.get(self.fxn, self.fxn(owner))
        self.__class__.CLASSPROP_CACHES[self.fxn] = result
        return self.__class__.CLASSPROP_CACHES[self.fxn]



""" pynchon.plugins.jinja
"""

from fleks import tagging

from pynchon import abcs, api, cli, models

from pynchon.util import files, lme, text, typing  # noqa

LOGGER = lme.get_logger(__name__)


@tagging.tags(click_aliases=["j"])
class Jinja(models.Planner):
    """Renders files with {jinja.template_includes}"""

    # diff --color --minimal -w --side-by-side /etc/bash.bashrc <(bash --pretty-print /etc/bash.bashrc )

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "jinja"
        template_includes: typing.List[str] = typing.Field(default=[])
        exclude_patterns: typing.List[str] = typing.Field()

        # @tagging.tagged_property(conflict_strategy="override")
        @property
        def exclude_patterns(self):
            from pynchon.config import globals

            # globals = plugin_util.get_plugin("globals").get_current_config()
            global_ex = globals.exclude_patterns
            my_ex = self.__dict__.get("exclude_patterns", [])
            return list(set(global_ex + my_ex + ["**/pynchon/templates/includes/**"]))

    name = "jinja"
    priority = 7
    COMMAND_TEMPLATE = (
        "python -mpynchon.util.text render jinja "
        "{src} --context-file {context_file} "
        "--output {output} {template_args}"
    )

    # cli_subsumes: typing.List[typing.Callable] = [
    #     # render_main.j2cli,
    #     # render_main.jinja_file,
    # ]

    def _get_jinja_context(self):
        """ """
        fname = ".tmp.jinja.ctx.json"
        with open(fname, "w") as fhandle:
            fhandle.write(text.to_json(self.project_config))
        return f"{fname}"

    @property
    def _include_folders(self):
        includes = self.project_config.jinja["template_includes"]
        from pynchon import api

        includes = api.render.get_jinja_includes(*includes)
        return includes

    @cli.click.flag("--local")
    def list_includes(
        self,
        local: bool = False,
    ):
        """Lists full path of each include-file

        :param local: bool:  (Default value = False)
        """
        includes = self._include_folders
        if local:
            includes.remove(api.render.PYNCHON_CORE_INCLUDES)
        includes = [abcs.Path(t) / "**/*.j2" for t in includes]
        LOGGER.warning(includes)
        matches = files.find_globs(includes)
        return matches

    @cli.click.flag("--local")
    def list_include_args(
        self,
        local: bool = False,
    ):
        """
        Lists all usable {% include ... %} values
        """
        includes = self.list_includes(local=local)
        out = []
        for fname in includes:
            fname = abcs.Path(fname)
            for inc in self._include_folders:
                try:
                    fname = fname.relative_to(inc)
                except ValueError:
                    continue
                else:
                    out.append(fname)
                break
            else:
                pass
        return out

    @tagging.tags(click_aliases=["ls"])
    def list(self, changes=False):
        """
        Lists affected resources for this project
        """
        default = self[:"project"]
        proj_conf = self[:"project.subproject":default]
        project_root = proj_conf.get("root", None) or self[:"git.root":"."]
        # project_root = proj_conf.get("root", None) or '.'
        globs = [
            abcs.Path(project_root).joinpath("**/*.j2"),
        ]
        self.logger.warning(f"search patterns are {globs}")
        result = files.find_globs(globs)
        self.logger.debug(f"found {len(result)} j2 files (pre-filter)")
        excludes = self["exclude_patterns"]
        self.logger.debug(f"filtering search with {len(excludes)} excludes")
        result = [p for p in result if not p.match_any_glob(excludes)]
        self.logger.debug(f"found {len(result)} j2 files (post-filter)")
        if not result:
            err = f"{self.__class__.__name__} is active, but found no .j2 files!"
            self.logger.critical(err)
        return result

    def plan(
        self,
        config=None,
    ) -> typing.List:
        """Creates a plan for this plugin"""

        def _get_template_args():
            """ """
            # import IPython; IPython.embed()
            templates = self["template_includes"]
            templates = [t for t in templates]
            templates = [f"--include {t}" for t in templates]
            templates = " ".join(templates)
            return templates

        plan = super(self.__class__, self).plan()
        jctx = self._get_jinja_context()
        templates = _get_template_args()
        for src in self.list():
            output = str(src).replace(".j2", "")
            plan.append(
                self.goal(
                    type="render",
                    resource=output,
                    command=self.COMMAND_TEMPLATE.format(
                        src=src,
                        context_file=jctx,
                        template_args=templates,
                        output=output,
                    ),
                )
            )
        return plan.finalize()



""" pynchon.models.python
"""

from fleks.models import BaseModel

from pynchon import abcs
from pynchon.util import lme, typing

LOGGER = lme.get_logger(__name__)


class EntrypointMetadata(BaseModel):
    is_click: bool = typing.Field(help="", required=False, default=False)
    is_package: bool = typing.Field(help="", required=False, default=False)
    is_module: bool = typing.Field(help="", required=False, default=True)
    bin_name: typing.StringMaybe = typing.Field(
        help="Name for this console script. (Nil if module-entrypoint)",
        required=False,
        default=None,
    )
    help_command: typing.StringMaybe = typing.Field(
        help="Command that returns help for this entrypoint",
        required=False,
        default=None,
    )
    dotpath: str = typing.Field(
        help="",
        required=True,
    )
    file: str = typing.Field(
        help="",
        required=True,
    )
    resource: abcs.Path = typing.Field(help="", required=True)
    path: abcs.Path = typing.Field(help="", required=True)
    entrypoints: typing.List = typing.Field(help="", required=False, default=[])
    src_root: abcs.Path = typing.Field(required=True)

    @property
    def src_url(self) -> str:
        return "/" + str(self.path.relative_to(self.src_root.parent))
        # return abcs.Path(path).absolute().relative_to(abcs.Path(git_root).absolute())

    @property
    def help_invocation(self):
        if self.help_command is not None:
            return self.help_command
        elif self.is_module:
            return f"python -m{self.dotpath} --help"
        # if self.is_package:
        #     return f""
        raise ValueError(f"Cannot determine help_invocation for {self}")

    @property
    def help_output(self):
        """
        Computes help-output given help-invocation.
        WARNING: obviously this has side-effects;
        if help-invocation is not safe to run, then it should be set to `None`
        """
        from pynchon.util.os import invoke

        if self.help_invocation:
            cmd = invoke(self.help_invocation)
            if cmd.succeeded:
                help = cmd.stdout.lstrip().strip()
                help = f"$ {self.help_invocation}\n\n{help}"
            else:
                LOGGER.critical(
                    f"ERROR: failure executing command (cannot extract help!)\n\ncommand='{self.help_invocation}'\n\nerror follows:\n\n{cmd.stderr}"
                )
                help = f"Failed to capture help from command `{self.help_invocation}`"
            return help
        else:
            import IPython

            IPython.embed()
            raise Exception(self)

    @property
    def module(self):
        return (
            f"{self.dotpath}.__main__"
            if str(self.file).endswith("__main__.py")
            else self.dotpath
        )



""" pynchon.constants
"""

import os
from pathlib import Path

GLYPH_COMPLEXITY = "🐉 Complex"

PYNCHON_ROOT = os.environ.get("PYNCHON_ROOT", None)
PYNCHON_CONFIG = os.environ.get("PYNCHON_CONFIG", None)
LOG_LEVEL = os.environ.get("PYNCHON_LOG_LEVEL", "WARNING")

CONF_FILE_SEARCH_ORDER = ["pynchon.json5", ".pynchon.json5", "pyproject.toml"]
DEFAULT_PLUGINS = [
    "core",
    "git",
    "pandoc",  # required by markdown
    "markdown",
    "src",
    "docs",
    "parse",
    "pattern",
    "plugins",
    "project",
    "globals",
    "github",  # NB: used by `pynchon pattern sync . github`
    "python",
    "gen",
    "render",
    "json",
    "jinja",
]
PYNCHON_EMBEDDED_TEMPLATES_ROOT = PETR = Path(__file__).parents[0] / "templates"
PYNCHON_CORE_INCLUDES_DIRS = (PYNCHON_EMBEDDED_TEMPLATES_ROOT / "includes",)
for _p in PYNCHON_CORE_INCLUDES_DIRS:
    assert _p.exists()

# TEMPLATE_DIR = os.environ.get(
#     "PYNCHON_TEMPLATE_DIR",
#     os.path.join(
#         os.path.dirname(__file__),
#         "templates",
#     ),
# )
#


# # FIXME: reuse parallel jinja env/template stuff in pynchon.util.text.render
# plugin_base = "pynchon/plugins"
# T_DETAIL_CLI = ENV.get_template(f"{plugin_base}/python/cli/detail.md.j2")
# T_TOC_CLI = ENV.get_template(f"{plugin_base}/python/cli/TOC.md.j2")
# T_VERSION_METADATA = ENV.get_template(f"{plugin_base}/core/VERSIONS.md.j2")
# T_CLI_MAIN_MODULE = ENV.get_template(f"{plugin_base}/python/cli/main.module.md.j2")



""" pynchon.util.events """

import functools
import collections

EventType = collections.namedtuple("Event", "type msg data")
LifeCycle = functools.partial(EventType, type="app-lifecycle")
ConfigFinalized = functools.partial(EventType, type="config-finalized")
PluginFinalized = functools.partial(EventType, type="plugin-finalized")


class Engine:
    """ """

    def push(self, **kwargs):
        event = EventType(**kwargs)
        raise NotImplementedError(event)

    def subscribe(self, fxn, type: str = None):
        """

        :param fxn: param type: str:  (Default value = None)
        :param type: str:  (Default value = None)

        """


DEFAULT_ENGINE = Engine()
push = DEFAULT_ENGINE.push



""" pynchon.util.text
    Utilities for parsing, generating or manipulating text
"""

from . import dumps, loadf, loads  # noqa

jsonify = to_json = dumps.json  # noqa


def indent(txt: str, level: int = 2) -> str:
    """
    indents text, or if given an object, stringifies and then indents

    :param txt: str:
    :param level: int:  (Default value = 2)
    """
    import pprint

    if not isinstance(txt, (str, bytes)):
        txt = pprint.pformat(txt)
    return "\n".join([(" " * level) + line for line in txt.split("\n") if line.strip()])



""" pynchon.plugins.pattern

    See also:
        https://github.com/cookiecutter/cookiecutter/issues/784
"""

import os

from fleks import cli, tagging

from pynchon import abcs, constants, models
from pynchon.api import render
from pynchon.util import lme, text, typing
from pynchon.util.os import invoke
from pynchon.util.files.diff import str_diff

LOGGER = lme.get_logger(__name__)

FNAME_ADVICE = ".scaffold.advice.json5"
PETR = abcs.Path(constants.PYNCHON_EMBEDDED_TEMPLATES_ROOT)


class RenderResult(abcs.Config):
    before: typing.Optional[str] = typing.Field(default=None)
    after: typing.Optional[str] = typing.Field(default=None)
    diff: typing.Optional[str] = typing.Field(default=None)
    src: typing.Union[str, abcs.Path] = typing.Field(required=True)
    dest: typing.Union[str, abcs.Path] = typing.Field(default=None)

    @property
    def diff(self):
        return str_diff(self.after, self.before)

    def __str__(self):
        return f"<RenderResult {self.src}>"


class ScaffoldAdvice(abcs.Config):
    file: typing.Union[str, abcs.Path] = typing.Field(required=True)
    loaded: bool = typing.Field(default=False)
    inherits: typing.List[str] = typing.Field(default=[])

    @property
    def inherits(self):
        if self.loaded:
            return self.__dict__["inherits"]
        else:
            return self.load().inherits

    def exists(self):
        return abcs.Path(self.file).exists()

    def load(self):
        cadvice = abcs.Path(self.file).read()
        adv = text.loads.json5(cadvice)
        self.__dict__.update(loaded=True, **adv)
        return self

    def __str__(self):
        return f"<ScaffoldAdvice {id(self)}>"


class Scaffold(abcs.Config):
    root: typing.Union[str, abcs.Path] = typing.Field(required=True)
    files: typing.List[str] = typing.Field(default=[])
    dirs: typing.List[str] = typing.Field(default=[])
    kind: str = typing.Field(required=True)

    def sync(self, goals=[], **kwargs):
        """ """
        # LOGGER.critical(f"sync: {kwargs}")
        self.sync_scaffold_dirs(goals=goals, **kwargs)
        self.sync_scaffold_files(goals=goals, **kwargs)
        for p in self.get_parents():
            p.sync(goals=goals, **kwargs)
        return goals

    def sync_scaffold_dirs(
        self,
        plugin=None,
        should_plan: bool = False,
        dest=None,
        goals=[],
        context: dict = {},
    ):
        """ """
        LOGGER.info(f"sync_scaffold_dirs: {dest}")
        for pdir in self.dirs:
            tmp = pdir.relative_to(self.root)
            if render.is_templated(str(tmp)):
                msg = f"detected `{tmp}` is templated, "
                tmp = abcs.Path(
                    render.get_template_from_string(str(tmp)).render(**context)
                )
                msg += f"rendered it as {tmp}"
                LOGGER.critical(msg)
            LOGGER.critical(f"sync_scaffold_dirs:: {dest} {tmp}")
            if tmp.exists():
                LOGGER.info(f"directory already exists @ `{tmp}`")
            else:
                goals.append(
                    models.Goal(
                        type="create",
                        label=f"dir-creation required by {self.kind}",
                        resource=tmp,
                        command=f"mkdir -p {tmp}",
                    )
                )

    def sync_scaffold_files(
        self,
        plugin=None,
        dest=None,
        goals=[],
        should_plan: bool = False,
        context: typing.Dict = {},
    ) -> models.Plan:
        """ """
        assert plugin
        # ctx_f = ".tmp.pattern.ctx"
        # dumpf.json(context, file=ctx_f)
        for src in self.files:
            dst = src.relative_to(self.root)
            plugin.render_file(
                goals=goals,
                kind=self.kind,
                context=context,
                src=src,
                dest=dst,
                should_plan=should_plan,
            )
            # Pattern.siblings['jinja'].render(
            #     src=src, dest=dest,
            #     should_plan=should_plan))
            # if not dst.exists():
            #     # goals.append(
            #     #     models.Goal(
            #     #         type="create",
            #     #         label=f"file-creation required by scaffold @ `{self.kind}`",
            #     #         resource=dst,
            #     #         command=f"cp {src} {dst}",
            #     #     )
            #     # )
            # else:
            #     LOGGER.critical(f"modifies?: {dest}")
            #     if should_plan:
            #         # goals.append(
            #         #     self.siblings['jinja'].render(
            #         #         src=src, dest=dest,
            #         #         should_plan=should_plan))
            #         goals.append(models.Goal(
            #         type='render',
            #         label=f"rendering template @ `{dst}`",
            #         resource=dst,
            #         command=f"pynchon jinja render {src} {dst}",
            #     ))

    def get_parents(self):
        """ """
        out = []
        parents = self.advice.inherits if self.advice else []
        for kind in parents:
            pfolder = PETR / "scaffolds" / kind
            scaf = Scaffold(kind=kind, root=pfolder)
            out.append(scaf)
        return out

    def get_inherited_files(self):
        """ """
        out = []
        for scaf in self.get_parents():
            out += scaf.files
        return out

    def get_inherited_dirs(self):
        """ """
        out = []
        for scaf in self.get_parents():
            out += scaf.dirs
        return out

    @property
    def has_advice(self) -> bool:
        return self.advice is not None

    @property
    def advice(self):
        fadvice = ScaffoldAdvice(file=self.root / FNAME_ADVICE)
        if not fadvice.exists():
            return None
        else:
            return fadvice

    @property
    def files(self) -> typing.List[str]:
        folder = self.root
        base_files = list(
            {x for x in folder.glob("*") if not x.is_dir() and x.name != FNAME_ADVICE}
        )
        return base_files  # +self.get_inherited_files()

    @property
    def dirs(self) -> typing.List:
        base_dirs = list(
            [x for x in self.root.glob("**/") if x.is_dir() and not x == self.root]
        )
        return base_dirs  # +self.get_inherited_dirs()

    def __str__(self):
        return f"<Scaffold@`{self.kind}`>"

    __repr__ = __str__


@tagging.tags(click_aliases=["pat"])
class Pattern(models.ResourceManager):
    """
    Tools for working with file/directory patterns
    """

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "pattern"
        include_patterns: typing.List[str] = typing.Field(default=["*/", "*/*/"])

        @property
        def root(self):
            tmp = PETR / "scaffolds"
            assert tmp.exists(), tmp
            return tmp

    name = "pattern"
    cli_name = "pattern"

    @property
    def patterns(self) -> typing.Dict:
        tmp = super(self.__class__, self).list()
        tmp = [[str(x.relative_to(self.config.root)), x.glob("**/*")] for x in tmp]
        tmp = dict(tmp)
        keep = [x for x in tmp if not any([k.startswith(f"{x}/") for k in tmp])]
        tmp = dict(list([[k, list(v)] for k, v in tmp.items() if k in keep]))
        return tmp

    @property
    def pattern_folder(self):
        return self["root"]

    @property
    def pattern_names(self):
        tmp = self.patterns
        tmp = [abcs.Path(x) for x in tmp.keys()]
        tmp = [x.parents[0] if (self.config.root / x).is_file() else x for x in tmp]
        tmp = map(str, tmp)
        return list(tmp)

    @tagging.tags(click_aliases=["ls"])
    def list(self) -> typing.List:
        """
        Describe templates we can cut patterns from
        """
        return self.pattern_names

    @cli.click.command("open")
    @cli.click.argument("kind", nargs=1)
    def _open(self, kind):
        """open pattern in editor"""
        pfolder = self.pattern_folder / kind
        ed = self[:"pynchon.editor":"atom"]
        invoke(f"{ed} {pfolder}&", system=True)

    @cli.options.plan
    @cli.click.option("--kind")
    @cli.click.argument("dest", nargs=1)
    @cli.click.argument("src", nargs=1)
    def render_file(
        self,
        kind: str = None,
        name: str = None,
        should_plan: bool = False,
        src: str = None,
        dest: str = None,
        goals: typing.List = [],
        context: typing.Dict = {},
    ):
        """Render a single file @ DEST from KIND"""
        pconf = Pattern.project_config.dict()
        context = context or dict(name=name or self[:"project.name":], **pconf)
        context.update(__template__=dest)
        pattern = Scaffold(kind=kind, root=self.pattern_folder / kind)
        dest = abcs.Path(dest)
        try:
            src_content = abcs.Path(src).read()
        except (Exception,) as exc:
            LOGGER.critical(f"error reading '{src}':\n\n{exc}")
            src_content = None
        src_templated = render.is_templated(src_content)
        if dest.exists():
            if src_templated:
                LOGGER.critical(f"render_file: {[src, dest]}")
                before = src_content
                tmpl = render.get_template_from_string(src_content)
                after = tmpl.render(**context)
                tmp = RenderResult(before=before, after=after, src=src, dest=dest)
                if tmp.diff:
                    LOGGER.critical("render_file: diff present")
                    LOGGER.critical(f"\n{tmp.diff}")
                    nsrc = abcs.Path(os.path.relpath(src, "."))
                    tmpf = dest.name.replace(os.path.sep, "_")
                    tmpf = abcs.Path(f".tmp.pattern.rendered.{tmpf}")
                    tmpf.write(after)
                    goal = models.Goal(
                        type="sync",
                        label="sync existing file",
                        resource=dest,
                        owner=Pattern.__name__,
                        udiff=tmp.diff,
                        command=f"cp {tmpf} {dest}",
                    )
                    goals.append(goal)
                else:
                    LOGGER.critical("render_file: no diff")
            else:
                LOGGER.critical("render_file: destination is missing (templated)")
        else:
            goals.append(
                models.Goal(
                    type="sync",
                    label="sync new file",
                    resource=dest,
                    command=f"cp '{src}' {dest}",
                )
            )
        if should_plan:
            return goals
        else:
            return self.apply(plan=goals)

    @cli.click.option("--name", default=None)
    @cli.click.argument("kind", nargs=1)
    @cli.click.argument("dest", nargs=1)
    @cli.options.plan
    def sync(
        self,
        dest: str = None,
        kind: str = None,
        name: str = None,
        should_plan: bool = False,
        context: typing.Dict = {},
    ):
        """Synchronizes DEST from scaffold KIND"""
        LOGGER.critical(f'Synchronizing "{dest}" from `{kind}`')
        tmp = self.pattern_names
        if kind not in tmp:
            LOGGER.critical(f"Unrecognized pattern `{kind}`; expected one of {tmp}")
            raise SystemExit(1)
        plan = super(self.__class__, self).plan()
        destp = abcs.Path(dest)

        pattern = Scaffold(kind=kind, root=self.pattern_folder / kind)

        LOGGER.warning(f"found pattern:\n\t{pattern}")
        LOGGER.warning(f"tentatively rendering {pattern} to `{destp}`")
        patterns = [pattern] + pattern.get_parents()
        folder = abcs.Path(dest).absolute()
        pconf = Pattern.project_config.dict()
        context = context or dict(name=name or self[:"project.name":], **pconf)
        for pattern in patterns:
            LOGGER.critical(f"running sync for: {pattern}")
            goals = pattern.sync(
                plugin=self,
                dest=destp,
                goals=plan,
                context=context,
                should_plan=should_plan,
            )
        if should_plan:
            LOGGER.critical(plan)
            return plan
        else:
            return self.apply(plan=plan)

    @cli.click.argument("name", nargs=1)
    @cli.click.argument("kind", nargs=1)
    @cli.options.plan
    def new(
        self,
        kind: str = None,
        name: str = None,
        should_plan: bool = False,
    ):
        """Instantiates PATTERN to NAME"""
        oname = name
        name = self[:"project.name":name]
        pfolder = self.pattern_folder / kind
        if not pfolder.exists():
            choices = set(self.list().keys())
            tmp = pfolder.relative_to(abcs.Path(".").absolute())
            LOGGER.critical(
                f'KIND @ "{name}" suggests pattern-folder @ "{tmp}", but folder does not exist!'
            )
            LOGGER.critical(f"Choices are: {choices}")
        else:
            plan = super(self.__class__, self).plan()
            dest = abcs.Path(name).absolute()
            if dest.exists():
                LOGGER.warning(f"{dest} already exists!")
            fadvice = ScaffoldAdvice(file=pfolder / ".scaffold.advice.json5")
            if fadvice.exists():
                fadvice.load()
                inherits = fadvice.inherits
            else:
                inherits = []
            inherits += [pfolder / "*"] if pfolder not in inherits else []
            LOGGER.warning(f"{kind} inherits {len(inherits)} patterns")
            dest = dest.relative_to(abcs.Path(".").absolute())
            dest2 = oname if oname == "." else dest
            # raise Exception(dest)
            for parent in inherits:
                # parent = parent.relative_to(abcs.Path(".").absolute())
                plan.append(
                    self.goal(
                        command=f"cp -rfv {parent} {dest2}",
                        resource=dest,
                        type="copy",
                    )
                )

            plan.append(
                self.goal(
                    command=f"{self.click_entry.name} {self.cli_name} render {dest}",
                    resource=dest.absolute(),
                    type="recursive-render",
                )
            )
            if should_plan:
                return plan
            else:
                result = self.apply(plan)
                return result.ok



""" pynchon.util.text.loads

    Helpers for loading data structures from strings
"""

import json as json_mod

import yaml as modyaml
import pyjson5 as modjson5

from pynchon.util import lme, typing

LOGGER = lme.get_logger(__name__)


def ini(content: str) -> typing.StringMaybe:
    """
    Parses `content` as ini-file.
    :param content: str:
    """
    raise NotImplementedError()


def yaml(content: str) -> typing.StringMaybe:
    """Parses `content` as yaml.

    :param content: str:
    """
    # try:
    return modyaml.safe_load(content)
    # except yaml.YAMLError as exc:
    #     print(exc)


def toml(content: str) -> typing.StringMaybe:
    """Parses `content` as toml.

    :param content: str:

    """
    raise NotImplementedError()


def json(content: str = "") -> typing.StringMaybe:
    """Parses `content` as JSON (strict).
    For most things, you're better using `loads.json5()`,
    since that's just a JSON superset with a more relaxed parser.

    :param content: str:  (Default value = '')
    """
    return json_mod.loads(content)


def json5(content: str = "", quiet=True) -> typing.Dict:
    """Parses `content` as JSON5.
    This tries to give a better error message than defaults.

    :param content: str:  (Default value = '')
    :param quiet: Default value = True)
    """
    try:
        return modjson5.loads(content)
    except (ValueError,) as exc:
        LOGGER.critical("Cannot parse json5 from literal!")
        quiet or LOGGER.critical(content)
        content_lines = content.split("\n")
        if "at column" in exc.args[0]:
            line_no = exc.args[0].split()[0].split(":")[-1]
            line_no = int(line_no)
            err_lines = "\n".join(content_lines[line_no - 3 : line_no + 3])
        else:
            err_lines = None
            line_no = None
        LOGGER.warning(f"error: {exc}")
        err_lines and LOGGER.warning(f"lines nearby:\n\n {err_lines}")
        raise



""" pynchon.models.planning
"""

import typing
import collections

import shil
from fleks import app, meta
from fleks.models import BaseModel

from pynchon import abcs

from pynchon.util import lme, typing  # noqa

RED_X = "❌"
RED_BALL = "🔴"
YELLOW_BALL = "🟡"


class BaseModel(BaseModel):
    @property
    def _action_summary(self):
        """ """
        if self.command:
            return shil.fmt(self.command)
        else:
            return f"{self.owner}.{self.callable.__name__}(..)"

    @property
    def rel_resource(self) -> str:
        return (
            abcs.Path(self.resource).absolute().relative_to(abcs.Path(".").absolute())
        )


class Goal(BaseModel):
    """ """

    # FIXME: validation-- command OR callable must be set

    class Config(BaseModel.Config):
        exclude: typing.Set[str] = {"udiff", "callable"}
        # arbitrary_types_allowed = True
        # json_encoders = {typing.MethodType: lambda c: str(c)}

    resource: abcs.ResourceType = typing.Field(default="?r", required=False)
    command: typing.StringMaybe = typing.Field(default=None)
    callable: typing.MethodType = typing.Field(default=None)
    type: typing.StringMaybe = typing.Field(default=None, required=False)
    owner: typing.StringMaybe = typing.Field(default=None)
    label: typing.StringMaybe = typing.Field(default=None)
    udiff: typing.StringMaybe = typing.Field(default=None)
    ordering: typing.StringMaybe = typing.Field(
        default=None,
        help="human-friendly string describing the sort order for this action inside plan",
    )

    def __rich__(self) -> str:
        """ """
        ordering = f" ({self.ordering.strip()})" if self.ordering else ""

        if self.udiff:
            sibs = [app.Markdown(f"```diff\n{self.udiff}\n```")]
        else:
            sibs = [
                app.Syntax(
                    f"  {self._action_summary}",
                    "bash",
                    line_numbers=False,
                    word_wrap=True,
                )
            ]

        return app.Panel(
            app.Group(*sibs),
            title=app.Text(f"{ordering} ", style="dim underline")
            + app.Text(f"{self.type}", style="dim bold yellow"),
            title_align="left",
            style=app.Style(
                dim=True,
                bgcolor="black",
                frame=False,
            ),
            subtitle=app.Text(f"{self.label or self.owner}", style="dim"),
            # + app.Text(" rsrc=", style="bold italic")
            # + app.Text(f"{self.rel_resource}", style="dim italic"),
        )


class Action(BaseModel):
    """ """

    type: str = typing.Field(default="unknown_action_type")
    ok: bool = typing.Field(default=None)
    error: str = typing.Field(default="")
    changed: bool = typing.Field(default=False)
    resource: abcs.ResourceType = typing.Field(default="??")
    command: str = typing.Field(default="echo")
    callable: typing.CallableMaybe = typing.Field(help="", default=None)
    owner: typing.StringMaybe = typing.Field(
        help="Name of the plugin that owns this Action", default=None
    )
    ordering: typing.StringMaybe = typing.Field(
        default=None,
        help="human-friendly string describing the sort order for this action inside plan",
    )

    def __rich__(self) -> str:
        """ """
        # indicator = RED_BALL if self.changed else YELLOW_BALL

        indicator = (
            app.Text(
                "modified",
                justify="right",
                style="red",
            )
            if self.changed
            else None
        )
        ind = (
            app.Text(
                "process: failed",
                # justify='right',
                style="red",
            )
            if not self.ok
            else None
        )
        err = (
            (
                app.Text(
                    "error: ",
                    # justify='right',
                    style="red",
                )
                + app.Text(self.error, justify="center")
            )
            if not self.ok
            else None
        )
        sibs = [
            app.Text(
                f"target: {self.rel_resource}",
            ),
            app.Text(f"action: {self._action_summary}"),
            indicator,
            ind,
            err,
        ]
        sibs = app.Group(*filter(None, sibs))
        ordering = f" ({self.ordering.strip()})"
        return app.Panel(
            sibs,
            title=app.Text(f"{ordering} ", style="dim underline")
            + app.Text(
                f"{self.type}", style=f"dim bold {'red' if self.changed else 'green'}"
            ),
            title_align="left",
            # style=app.Style(
            #     dim=True,
            #     # color='green',
            #     bgcolor="black",
            #     frame=False,
            # ,
            subtitle=app.Text(f"{self.owner}", style="dim"),
        )

    @property
    def status_string(self):
        if self.ok is None:
            tmp = "pending"
        elif self.ok:
            tmp = "ok"
        else:
            tmp = "failed"
        return tmp

    def __str__(self):
        return f"<[{self.type}]@{self.resource}: {self.status_string}>"


class Plan(typing.BaseModel):
    """ """

    goals: typing.List[Goal] = typing.Field(default=[])

    def finalize(self):
        """
        When a plan is finished being appended to,
        it can be "finalized" to set the `ordering` value
        for individual goals.
        """
        plan_length = len(self.goals)
        self.goals = [
            g.copy(update=dict(ordering=f"{i+1}/{plan_length}"))
            for i, g in enumerate(self.goals)
        ]
        return self

    def __rich__(self) -> str:
        """ """
        syntaxes = []
        for g in self.goals:
            if hasattr(g, "__rich__"):
                syntaxes.append(g.__rich__())
            else:
                syntaxes.append(str(g))

        table = app.Table.grid(
            # title=f'{__name__} ({len(self)} items)',
            # subtitle='...',
            # box=box.MINIMAL_DOUBLE_HEAD,
            expand=True,
            # border_style='dim italic yellow'
            # border_style='bold dim',
        )
        [
            [
                table.add_row(x),
                # table.add_row(app.Align(app.Emoji("gear"), align='right')),
            ]
            for i, x in enumerate(syntaxes)
        ]

        panel = app.Panel(
            table,
            title=app.Text(
                f"{self.__class__.__name__}", justify="left", style="italic"
            ),
            title_align="left",
            padding=1,
            style=app.Style(
                dim=True,
                # color='green',
                bgcolor="black",
                frame=False,
            ),
            subtitle=f"(Planned {len(self)} items)",  # subtitle=Text("✔", style='green')
        )
        return panel

    def append(self, other: Goal):
        """ """
        if other in self:
            return
        elif isinstance(other, (Goal,)):
            self.goals += [other]
        elif isinstance(other, (Plan,)):
            self.goals += other.dict()["goals"]
        elif isinstance(
            other,
            (
                list,
                # tuple,
            ),
        ):
            self.goals += other
        else:
            raise NotImplementedError(type(other))

    def __contains__(self, g):
        return g in self.goals

    def __len__(self):
        return len(self.goals)

    def __add__(self, other):
        """ """
        if isinstance(other, (Goal,)):
            return Plan(goals=self.goals + [other])
        elif isinstance(other, (Plan,)):
            return Plan(goals=self.goals + other.goals)
        elif isinstance(
            other,
            (
                list,
                tuple,
            ),
        ):
            return Plan(goals=self.goals + list(other))
        else:
            raise NotImplementedError(type(other))

    __iadd__ = __add__


class ApplyResults(typing.List[Action], metaclass=meta.namespace):
    """ """

    @property
    def ok(self):
        return all([a.ok for a in self])

    @property
    def action_types(self):
        tmp = list({g.type for g in self})
        return {k: [] for k in tmp}

    @property
    def _dict(self):
        """ """
        result = collections.OrderedDict()
        result["ok"] = self.ok
        result["resources"] = list({a.resource for a in self})
        result["actions"] = [
            g.command if g.command else self.callable.__name__ for g in self
        ]
        result["action_types"] = self.action_types
        result["changed"] = list({a.resource for a in self if a.changed})
        for g in self:
            result["action_types"][g.type].append(g.resource)
        return result

    def __str__(self):
        return f"<{self.__class__.__name__}[{len(self)} actions]>"



""" pynchon.plugins.python.cli
"""

import glob
import importlib

import shimport
from fleks import cli, tagging
from fleks.util.click import click_recursive_help
from memoized_property import memoized_property

from pynchon import abcs, api, models
from pynchon.models.python import EntrypointMetadata

from pynchon.util import lme, typing  # noqa

config_mod = shimport.lazy(
    "pynchon.config",
)
LOGGER = lme.get_logger(__name__)
import click


def _check_click(fxn=None, path=None) -> bool:
    """ """
    from pynchon.util.oop import is_subclass

    if path is not None:
        with open(str(path)) as fhandle:
            return "click" in fhandle.read()
    else:
        return any(
            [isinstance(fxn, (x,)) for x in [click.Group, click.Command]]
        ) or any([is_subclass(fxn, x) for x in [click.Group, click.Command]])


class PythonCliConfig(abcs.Config):
    """ """

    config_key: typing.ClassVar[str] = "python_cli"
    src_root: str = typing.Field(help="")
    # entrypoints: typing.List[typing.Dict] = typing.Field(help="")
    hooks: typing.List[str] = typing.Field(
        help="applicable hook names",
        default=["open-after-apply"],
    )

    @property
    def root(self):
        tmp = self.__dict__.get("root")
        if tmp:
            return tmp
        else:
            from pynchon import config

            return abcs.Path(config.docs.root) / "cli"

    @property
    def src_root(self) -> abcs.Path:
        """ """
        src_root = config_mod.src["root"]
        # FIXME: support for subprojects
        # # src_root = abcs.Path(
        # # config_mod.project.get(
        # #     "src_root", config_mod.pynchon.get("src_root")
        # # )).absolute()
        return abcs.Path(src_root)

    @property
    def module_entrypoints(self) -> typing.List[typing.Dict]:
        """ """
        src_root = self.src_root
        pat = src_root / "**" / "__main__.py"
        excludes = config_mod.src["exclude_patterns"]
        matches = glob.glob(str(pat), recursive=True)
        LOGGER.info(f"{len(matches)} matches for `entrypoints` filter")
        # LOGGER.info(f"filtering with `excludes`: {excludes}")
        matches = list(
            filter(lambda x: not abcs.Path(x).match_any_glob(excludes), matches)
        )
        # LOGGER.info(f"{len(matches)} matches survived filter")
        matches = [[x, {}] for x in matches]
        matches = dict(matches)
        pkg_name = (
            "unknown"  # self.siblings['python']['package'].get("name") or "unknown"
        )
        for f, meta in matches.items():
            LOGGER.info(f"found entry-point: {f}")
            dotpath = abcs.Path(f).relative_to(src_root)
            dotpath = ".".join(str(dotpath).split("/")[:-1])
            rsrc = self.root / f"{dotpath}.md"
            matches[f] = {
                **matches[f],
                **dict(
                    click=_check_click(path=f),
                    dotpath=dotpath,
                    file=f,
                    path=abcs.Path(f),
                    main_entrypoint=f.endswith("__main__.py"),
                    # package_entrypoint=False,
                    resource=rsrc,
                ),
            }
        result = list(matches.values())
        result = [EntrypointMetadata(src_root=self.src_root, **x) for x in result]
        return result


@tagging.tags(click_aliases=["pc"])
class PythonCLI(models.Planner):
    """Generators for Python CLI docs"""

    name = "python-cli"
    cli_name = "python-cli"
    config_class = PythonCliConfig

    @cli.click.group
    def gen(self):
        """Generates CLI docs for python packages"""

    @cli.click.flag("--changes")
    def list(self, changes: bool = False) -> typing.List[str]:
        """list related targets/resources"""
        if changes:
            out = []
            git = self.siblings["git"]
            git_changes = git.list(changes=True)
            for emeta in self.config.module_entrypoints:
                p = abcs.Path(emeta["path"]).absolute()
                if p in git_changes:
                    out.append(p)
            return out
        else:
            return [
                abcs.Path(emeta["path"]).absolute()
                for emeta in self.config.module_entrypoints
            ]

    @memoized_property
    def console_script_entrypoints(self) -> typing.List[EntrypointMetadata]:
        """ """
        python_platform_config = self.siblings["python"].config
        package_entrypoints = python_platform_config["package"]["console_scripts"]
        out = []
        for dct in package_entrypoints:
            bin_name = dct["bin_name"]
            module_name, fxn_name = dct["setuptools_entrypoint"].split(":")
            dotpath = ".".join([module_name, fxn_name])
            module = shimport.import_module(module_name)
            fxn = getattr(module, fxn_name, None)
            file = module.__file__
            emd = EntrypointMetadata(
                is_click=_check_click(fxn=fxn),
                is_package=True,
                file=file,
                bin_name=bin_name,
                help_command=f"{bin_name} --help",
                path=abcs.Path(file),
                resource=self.root / f"script-{dotpath}.md",
                src_root=self.src_root,
                is_module=False,
                dotpath=dotpath,
            )
            out.append(emd)
        return out

    @gen.command("toc")
    @cli.options.header
    @cli.options.output
    def toc(
        self,
        # format, file, stdout,
        output,
        header,
    ) -> None:
        """
        Generate table-of-contents for all project entrypoints
        """
        output = output or self.root / "README.md"
        LOGGER.warning(f"writing toc to file: {output}")
        entrypoints = self.console_script_entrypoints + self.config.module_entrypoints
        cfg = {**self.config.dict(), **dict(entrypoints=entrypoints)}
        cfg = {**api.project.get_config().dict(), **{self.config_class.config_key: cfg}}
        templatef = self.plugin_templates_root / "TOC.md.j2"
        tmpl = api.render.get_template(templatef)
        result = tmpl.render(
            # package_entrypoints=python_cli.entrypoints,
            package_entrypoints=[e for e in entrypoints if e.is_package],
            main_entrypoints=[e for e in entrypoints if e.is_module],
            **cfg,
        )
        with open(str(output), "w") as fhandle:
            fhandle.write(result)

    def _click_recursive_help(
        self, resource=None, path=None, module=None, dotpath=None, name=None, **kwargs
    ):
        """ """

        result = []
        if name and not module:
            module, name = name.split(":")
        if module and name:
            try:
                mod = importlib.import_module(module)
                entrypoint = getattr(mod, name)
            except (Exception,) as exc:
                LOGGER.critical(exc)
                return []
        else:
            err = "No entrypoint found"
            LOGGER.warning(err)
            raise Exception(err)
        LOGGER.debug(f"Recursive help for `{module}:{name}`")
        # raise Exception(dir())
        result = click_recursive_help(
            entrypoint,
            parent=None,
            path=path,
            dotpath=dotpath,
            module=module,
        ).values()
        git_root = self.siblings["git"]["root"]
        result = [
            {
                **v,
                **dict(
                    module=module,
                    resource=resource or self.root / f"{v['dotpath']}.md",
                    package=module.split(".")[0],
                    entrypoint=name,
                    dotpath=dotpath,
                    help="???",
                    # help=shil.invoke(
                    #     f"python -m{v['help_invocation']} --help", strict=True
                    # ).stdout,
                    # src_url=self.get_src_url(path),
                ),
            }
            for v in result
        ]
        return result

    #     """
    #     Generates help for every entrypoint
    #     """
    #     conf = util.python.load_entrypoints(util.python.load_setupcfg(path=file))
    #     entrypoints = conf.get("entrypoints", {})
    #     if not entrypoints:
    #         LOGGER.warning(f"failed loading entrypoints from {file}")
    #         return []
    #     docs = {}
    #     for e in entrypoints:
    #         bin_name = str(e["bin_name"])
    #         epoint = e["setuptools_entrypoint"]
    #         fname = os.path.join(output_dir, bin_name)
    #         fname = f"{fname}.md"
    #         LOGGER.debug(f"{epoint}: -> `{fname}`")
    #         docs[fname] = {**_click_recursive_help(name=e["setuptools_entrypoint"]), **e}
    #
    #     for fname in docs:
    #         with open(fname, "w") as fhandle:
    #             fhandle.write(constants.T_DETAIL_CLI.render(docs[fname]))
    #         LOGGER.debug(f"wrote: {fname}")
    #     return list(docs.keys())

    def get_entrypoint_metadata(
        self, file: str = None, console_script: bool = False
    ) -> EntrypointMetadata:
        """ """
        if console_script:
            LOGGER.warning(f"looking up console-script metadata for '{file}'")
            filtered = self.console_script_entrypoints
        else:
            LOGGER.warning(f"looking up module-entrypoint metadata for '{file}'")
            filtered = self.config["module_entrypoints"]
        found = False
        file = abcs.Path(file)
        for emd in filtered:
            if str(emd.path) == str(file):
                if console_script:
                    raise Exception(emd.dict())
                else:
                    module = shimport.import_module(emd.module)
                    wrapped = shimport.wrapper(module)
                    click_entries = wrapped.filter(
                        only_functions=True, filter_vals=[_check_click]
                    )
                    click_entry = (
                        list(click_entries.items())[0] if click_entries else None
                    )
                    if click_entry is None:
                        LOGGER.critical(
                            f"exception retrieving help programmatically: {file}"
                        )
                        # LOGGER.critical(
                        #     f"error retrieving help via system CLI: {emd.help_invocation}"
                        # )
                        recursive_help = {}
                    else:
                        name, fxn = click_entry
                        recursive_help = self._click_recursive_help(
                            module=emd.module,
                            name="entry",
                            # resource=self.root / f"{emd.dotpath}.md",
                            resource=emd.resource,
                            dotpath=emd.dotpath,
                            path=emd.path,
                            file=emd.file,
                        )
                # try:
                # emd.update(entrypoints=tmp)
                # except (AttributeError,) as exc:
                # else:
                # rsrc = self.root / f"{emd.dotpath}.md"
                # docs_url = rsrc.relative_to(self.docs_root.parent)
                # emd.update(
                #     is_click=True,
                #     # help_invocation=help_invocation,
                #     docs_url=docs_url,
                #     # src_url='/'+str(src_url),
                #     # src_url=src_url,
                #     resource=rsrc,
                #     # entrypoints=sub_entrypoints,
                # )
                # metadata.update(**get_cmd_output(help_invocation))
                # import IPython; IPython.embed()
                # raise Exception(sub_entrypoints)
                found = True
                break
        if not found:
            LOGGER.critical(f"missing {file}")
            return {}
        return emd

    @property
    def docs_root(self):
        return abcs.Path(self[:"docs.root":])

    @memoized_property
    def src_root(self):
        return abcs.Path(self[:"src.root":])

    @property
    def root(self):
        return self.config.root

    @gen.command("module-doc-gen")
    @cli.options.stdout
    @cli.options.file
    @cli.options.header
    @cli.options.output_file
    def module_doc_gen(
        self,
        file,
        output,
        stdout,
        header,
    ):  # noqa
        """
        Autogenenerate docs for py modules using `__main__`
        """
        assert abcs.Path(file).exists(), f"input file @ {file} does not exist"
        metadata = self.get_entrypoint_metadata(file=file)
        output = abcs.Path(output) if output else self.root / f"{metadata.dotpath}.md"
        output_dir = output.parents[0]
        assert output_dir.exists(), f"{output_dir} does not exist"
        tmpl = api.render.get_template(self.plugin_templates_root / "main.module.md.j2")
        config = {
            **api.project.get_config().dict(),
            **{self.config_class.config_key: {**self.config.dict(), **dict()}},
        }
        result = tmpl.render(
            entrypoints=[metadata],
            **config,
        )
        LOGGER.critical(result)
        LOGGER.critical(f"Writing output to: {output}")
        with open(str(output), "w") as fhandle:
            fhandle.write(result)

    @gen.command("script-doc-gen")
    @cli.options.file
    @cli.options.output_file
    def script_doc_gen(self, file, output):
        """
        Autogenenerates docs for py files mentioned in `console_scripts`.
        (Not for direct usage; let the planner determine invocation)
        """
        assert abcs.Path(file).exists(), f"input file @ {file} does not exist"
        metadata = self.get_entrypoint_metadata(file, console_script=True)
        output = (
            abcs.Path(output) if output else self.root / f"script-{metadata.dotpath}.md"
        )
        output_dir = output.parents[0]
        assert output_dir.exists(), f"{output_dir} does not exist"
        tmpl_f = self.plugin_templates_root / "script.console.md.j2"
        tmpl = api.render.get_template(tmpl_f)
        result = tmpl.render(entrypoints=[metadata])
        LOGGER.critical(result)
        LOGGER.critical(f"Writing output to: {output}")
        with open(str(output), "w") as fhandle:
            fhandle.write(result)

    @property
    def plugin_invocation(self):
        return f"{self.click_entry.name} {self.cli_name}"

    def plan(self):
        """Describe plan for this plugin"""
        plan = super(self.__class__, self).plan()

        plan.append(
            self.goal(command=f"mkdir -p {self.root}", type="mkdir", resource=self.root)
        )

        rsrc = self.root / "README.md"
        cmd = f"{self.plugin_invocation} toc " f"--output {rsrc}"
        plan.append(self.goal(command=cmd, type="gen", resource=rsrc))

        script_entrypoints = self.console_script_entrypoints
        module_entrypoints = self.config.module_entrypoints
        if not script_entrypoints:
            LOGGER.warning("no script_entrypoints found..")
        if not module_entrypoints:
            LOGGER.warning("no module_entrypoints found..")
        for entrypoint_metadata in script_entrypoints:
            rsrc = entrypoint_metadata.resource
            plan.append(
                self.goal(
                    command=(
                        f"{self.plugin_invocation} script-doc-gen "
                        f"--file {entrypoint_metadata.file} "
                        f"--output {rsrc} "
                    ),
                    type="gen",
                    resource=rsrc,
                )
            )
        for entrypoint_metadata in module_entrypoints:
            entrypoint_metadata = self.get_entrypoint_metadata(entrypoint_metadata.file)
            inp = entrypoint_metadata.path
            rsrc = entrypoint_metadata.resource
            if not rsrc:
                raise Exception(entrypoint_metadata)
            plan.append(
                self.goal(
                    command=(
                        f"{self.plugin_invocation} module-doc-gen "
                        f"--file {inp} --output {rsrc}"
                    ),
                    type="gen",
                    resource=rsrc,
                )
            )
        return plan.finalize()



""" {{pkg}}.util.types

This module collects common imports and annotation-types, i.e.
various optional/composite types used in type-hints, underneath
one convenient namespace.
"""

import typing
from pathlib import Path as BasePath

from fleks.util.typing import *  # noqa

validate = validate_arguments


def bind_method(func, instance, as_name=None):
    """
    Binds the function *func* to *instance*, with either provided name *as_name*
    or the existing name of *func*. The provided *func* should accept the
    instance as the first argument, i.e. "self".

    :param instance: param func:
    :param as_name: Default value = None)
    :param func:
    """
    assert isinstance(func, (FunctionType,))
    if as_name is None:
        as_name = func.__name__
    bound_method = func.__get__(instance, instance.__class__)
    setattr(instance, as_name, bound_method)
    return bound_method


OptionalAny = typing.Optional[typing.Any]
PathType = type(BasePath())

Bool = bool
NoneType = type(None)
BoolMaybe = typing.Optional[bool]
StringMaybe = typing.Optional[str]
CallableMaybe = typing.Optional[typing.Callable]
DictMaybe = typing.Optional[typing.Dict]
TagDict = typing.Dict[str, str]


Namespace = typing.Dict[str, typing.Any]
CallableNamespace = typing.Dict[str, typing.Callable]

# i.e. `obj,created = model.objects.get_or_create()`
GetOrCreateResult = typing.Tuple[object, bool]



""" pynchon.codemod.docstrings.javadoc """

import libcst as cst

from pynchon.util import lme

from .base import base

LOGGER = lme.get_logger(__name__)


class module(base):
    DESCRIPTION: str = """\n\tAdds or updates javadoc for Modules"""


class function(base):
    DESCRIPTION: str = """\n\tAdds or updates javadoc for Functions"""

    def leave_FunctionDef(
        self,
        original_node: cst.FunctionDef,
        updated_node: cst.FunctionDef,
    ) -> cst.FunctionDef:
        # LOGGER.critical(original_node)
        return original_node



""" pynchon.plugins.markdown
"""

import marko
from fleks import tagging
from marko.ast_renderer import ASTRenderer

from pynchon import abcs, api, cli, events, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)

ElementList = typing.List[typing.Dict]
# from pynchon.plugins.tests import DocTestSuite
# markdown_suite = DocTestSuite(
#     suite_name="markdown",
# )


class Markdown(models.CliPlugin):
    """Markdown"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "markdown"
        goals: typing.List[str] = typing.Field(default=[])
        include_patterns: typing.List[str] = typing.Field(default=[])
        exclude_patterns: typing.List[str] = typing.Field(default=[])
        root: typing.Union[str, abcs.Path, None] = typing.Field(default=None)
        # tests: DocTestSuite = typing.Field(default=markdown_suite)

    name = "markdown"
    # @cli.click.flag("-p", "--python", help="only python codeblocks")
    cli_name = "markdown"
    priority = 0

    @cli.click.flag("-p", "--python", help="only python codeblocks")
    @cli.click.flag("-b", "--bash", help="only bash codeblocks")
    @cli.click.argument("file")
    def doctest(
        self,
        file: str = None,
        python: bool = False,
        bash: bool = False,
    ) -> ElementList:
        assert python or bash
        element_lst = self.parse(file=file, python=python, bash=bash)
        if not element_lst:
            LOGGER.critical(f"filtered element list is empty! {element_lst}")

        def _doctest(element):
            LOGGER.critical(element)
            child = element["children"][0]
            assert child["element"] == "raw_text"
            script: str = child["children"]
            raise Exception(script)
            # return #shil.invoke(script,...))

        for el in element_lst:
            el.update(_doctest(el))
        return element_lst

    @tagging.tags(click_aliases=["parse.markdown"])
    @cli.click.flag("-c", "--codeblocks", help="only codeblocks")
    @cli.click.flag("-p", "--python", help="only python codeblocks")
    @cli.click.flag("-b", "--bash", help="only bash codeblocks")
    @cli.click.argument("file")
    def parse(
        self,
        file: str = None,
        codeblocks: bool = False,
        python: bool = False,
        bash: bool = False,
    ) -> ElementList:
        """parses given markdown file into json"""
        codeblocks = codeblocks or python or bash
        assert file
        with open(file) as fhandle:
            content = fhandle.read()

        parsed = marko.Markdown(renderer=ASTRenderer)(content)
        # def walk(thing):
        #     LOGGER.critical(thing)
        #     if isinstance(thing, (dict,)):
        #         children = thing.pop('children', [])
        #         if isinstance(children,(str,)):
        #             return children
        #         if not children:
        #             return thing
        #         else:
        #             return [walk(ch) for ch in children]
        #     if isinstance(thing, (list,)):
        #         return [walk(x) for x in thing]
        #     else:
        #         return thing
        #
        children = parsed["children"]
        out = []
        for child in children:
            if child.get("element") == "fenced_code":
                lang = child.get("lang")
                if lang is not None:
                    out.append(child)
        LOGGER.critical(child)
        if python:
            out = [ch for ch in out if child.get("lang") == "python"]
        if bash:
            out = [ch for ch in out if child.get("lang") == "bash"]
        # out=[]
        # for child in children:
        #     if bash and lang=='bash':
        #         child['code'] = ''.join([
        #             x['children'] for x in tmp ])
        #         out.append(child)
        #     elif python and lang=='python':
        #         child['code'] = ''.join([
        #             x['children'] for x in tmp ])
        #         out.append(child)

        return out
        # for child in children:
        #     result import pydash
        # flat = pydash.flatten_deep(children)
        # flat = [pydash.flatten_deep(x) for x in flat]
        # if codeblocks:
        #     result = [x for x in flat if x.get("element") == "fenced_code"]
        # if python:
        #     assert not bash
        #     result = [x for x in result if x.get("lang") == "python"]
        # if bash:
        #     assert not python
        #     result = [x for x in result if x.get("lang") == "bash"]
        # import IPython; IPython.embed()
        # return result

    # plan=None
    # def plan(self, config=None):
    #     """Describe plan for this plugin"""
    #     plan = super().plan(config=config)
    #     return plan
    # resources = [abcs.Path(fsrc) for fsrc in self.list()]
    # self.logger.warning("Adding user-provided goals")
    # for g in self["goals"]:
    #     plan.append(self.goal(command=g, resource="?", type="user-config"))
    #
    # self.logger.warning("Adding file-header related goals")
    # cmd_t = "python -mpynchon.util.files prepend --clean "
    # loop = self._get_missing_headers(resources)
    # for rsrc in loop["files"]:
    #     if rsrc.match_any_glob(self["exclude_patterns"::[]]):
    #         continue
    #     ext = rsrc.full_extension()
    #     ext = ext[1:] if ext.startswith(".") else ext
    #     # fhdr = header_files[ext]
    #     fhdr = self._render_header_file(rsrc)
    #     plan.append(
    #         self.goal(
    #             resource=rsrc,
    #             type="change",
    #             label=f"Adding file header for '{ext}'",
    #             command=f"{cmd_t} {fhdr} {rsrc}",
    #         )
    #     )



""" pynchon.app
"""

import sys
import atexit

import enlighten
from fleks import app as fleks_app
from fleks.app import AppBase, Console, Text, Theme
from memoized_property import memoized_property

from pynchon import events
from pynchon.util import lme

# from fleks.app import (AppBase, AppEvents)

LOGGER = lme.get_logger(__name__)


class AppEvents(AppBase):
    def __init__(self, **kwargs):
        """ """
        self.events = events


class AppConsole(fleks_app.AppBase):
    Text = Text
    Theme = Theme
    # docs = manager.term.link(
    #     'https://python-enlighten.readthedocs.io/en/stable/examples.html',
    #     'Read the Docs')

    def __init__(self, **kwargs):
        """ """
        self.console = Console()

    # # FIXME: use multi-dispatch over kwargs and define `lifecyle` repeatedly
    # def lifecycle_stage(self, sender, stage=None, **kwargs):
    #     """ """
    #     if stage:
    #         tmp = getattr(sender, '__name__', str(sender))
    #         # LOGGER.critical(f"STAGE ({tmp}): {stage}")
    #         self.status_bar.update(stage=stage)

    @memoized_property
    def status_bar(self):
        """ """
        tmp = self.manager.status_bar(
            status_format="{app}{fill}{stage}{fill}{elapsed}",
            color="bold_underline_bright_white_on_lightslategray",
            justify=enlighten.Justify.LEFT,
            app="Pynchon",
            stage="...",
            autorefresh=True,
            min_delta=0.1,
        )

        # atexit.register(
        #     lambda: self.events.lifecycle.send(
        #         self, stage=r"\o/" if not self.exc else "❌", msg=""
        #     )
        # )  # noqa: W605
        return tmp

    #
    # @memoized_property
    # def lifecycle_bar(self):
    #     """ """
    #     tmp = self.manager.status_bar(
    #         status_format=u'{fill}{msg}{fill}',
    #         color='bold_underline_bright_red_on_lightslategray',
    #         justify=enlighten.Justify.CENTER,
    #         msg='222',
    #         autorefresh=True,
    #         min_delta=0.1,
    #     )
    #
    #     atexit.register(
    #         lambda: self.events.lifecycle.send(self, msg="\o/")
    #     )  # noqa: W605
    #     return tmp

    @memoized_property
    def manager(self):
        tmp = enlighten.get_manager()
        atexit.register(lambda: self.manager.stop())
        return tmp


class AppExitHooks(fleks_app.AppBase):
    """ """

    exc = None
    hooks_installed = False
    # https://stackoverflow.com/questions/9741351/how-to-find-exit-code-or-reason-when-atexit-callback-is-called-in-python

    # def uninstall(self):
    def install_exit_hooks(self) -> None:
        if not self.hooks_installed:
            msg = "Installing exit handlers"
            LOGGER.critical(msg)
            self.events.lifecycle.send(self, msg=msg)
            self._sys_exit = sys.exit
            self._orig_exc_handler = sys.excepthook
            sys.exit = self.exit
            sys.excepthook = self.exc_handler
            atexit.register(self.exit_handler)
            self.hooks_installed = True

    def exit(self, code=0):
        self.exit_code = code
        self._sys_exit(code)

    def exc_handler(self, exc_type, exc, *args):
        self.exception = exc
        # import sys
        # ex_type, ex_value, traceback = sys.exc_info()
        # raise ex_type(ex_value)
        tmp = f"death by exc: ({self.exc})"
        self.events.lifecycle.send(self, stage=tmp)
        return self._orig_exc_handler(exc_type, exc, *args)

    def sys_exit_handler(self):
        if self.exit_code is not None and not self.exit_code == 0:
            tmp = f"death by sys.exit({self.exit_code})"
            self.events.lifecycle.send(self, stage=tmp)
            return True

    def exit_handler(self):
        """ """
        handled = self.sys_exit_handler()
        handled = handled or self.exc_exit_handler()
        handled = handled or self.default_exit_handler()
        return handled

    def exc_exit_handler(self):
        """ """
        if self.exception is not None:
            text = f"exc_exit_handler: {self.exception}"
            LOGGER.critical(text)
            text = self.Text(text)
            text.stylize("bold red", 0, 6)
            self.console.print(text)
            self.events.lifecycle.send(self, stage="❌")
            return True

    def default_exit_handler(self):
        """ """
        exc = self.exception
        if exc:
            LOGGER.critical(f"default_exit_handler: encountered exception: {exc}")
            self.events.lifecycle.send(self, stage="❌")
        else:
            pass
            # LOGGER.critical(f"default_exit_handler: "+r"\o/")
            # self.events.lifecycle.send(self, stage=r"\o/")
            # self.status_bar.update(stage=r"\o/", msg="", force=True)
            # import IPython; IPython.embed()
            # LOGGER.critical("exit ok")
        return True


class App(AppConsole, AppEvents, AppExitHooks):
    def __init__(self, **kwargs):
        """ """
        AppConsole.__init__(self, **kwargs)
        AppEvents.__init__(self, **kwargs)
        self.exit_code = None
        self.exception = None
        self.install_exit_hooks()
        # self.logger = ..


app = App()



""" pynchon.plugins.griffe
"""

from fleks import tagging

from pynchon import abcs, cli, events, models  # noqa
from pynchon.util import lme, os, typing  # noqa

LOGGER = lme.get_logger(__name__)


@tagging.tags(click_aliases=["g", "gr"])
class Griffe(models.ToolPlugin):
    """Tools for working with Python ASTs"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "griffe"

    name = "griffe"
    cli_name = "griffe"

    @tagging.tags(click_aliases=["dump"])
    @cli.click.option(
        "--package", "-p", "pkg", help="dotpath for a python package to use", default=""
    )
    def list(self, pkg: str = ""):
        """dump package signature from griffe"""
        pkg = pkg or self[:"python.package.name":]
        # return invoke(f'griffe {pkg}',load_json=True,log_command=True)
        # invoke(f"griffe {pkg} > .tmp.griffe")
        # from pynchon.util.text import loadf
        # return loadf.json(".tmp.griffe")
        return os.slurp_json(
            f"griffe {pkg}",
        )

    @cli.click.option(
        "--classes", "-c", is_flag=True, default=False, help="return classes"
    )
    @cli.click.option(
        "--functions", "-f", is_flag=True, default=False, help="return classes"
    )
    @cli.click.option(
        "--modules", "-m", is_flag=True, default=False, help="return classes"
    )
    def grep(
        self, classes: bool = False, functions: bool = False, modules: bool = False
    ):
        """grep for types"""
        result = self.list()
        print(result)
        raise NotImplementedError()
        # return result



""" pynchon.util.complexity
"""

import os
import ast
import sys
from collections import OrderedDict

import griffe
import mccabe

from pynchon import annotate, constants
from pynchon.abcs import Path
from pynchon.util import lme

WORKING_DIR = Path(".")
LOGGER = lme.get_logger(__name__)

GLYPH_COMPLEXITY = constants.GLYPH_COMPLEXITY


def clean_text(txt: str) -> str:
    """

    :param txt: str:
    :param txt: str:

    """
    return "\n".join([line for line in txt.split("\n") if line.strip()])


def get_module(package: str = "", file: str = ""):
    """

    :param package: str:  (Default value = "")
    :param file: str:  (Default value = "")
    :param package: str:  (Default value = "")
    :param file: str:  (Default value = "")

    """
    if not bool(package) ^ bool(file):
        err = "Expected --file or --package, but not both"
        raise RuntimeError(err)
    if file:
        file = os.path.abspath(file)
        new_path = os.path.dirname(file)
        assert os.path.exists(file)
        LOGGER.warning(f"modifying sys.path to include {new_path}")
        sys.path.append(new_path)
        package = os.path.splitext(os.path.basename(file))[0]
        working_dir = os.path.dirname(file)
    else:
        working_dir = WORKING_DIR
    loader = griffe.loader.GriffeLoader()
    module = loader.load_module(package)
    annotate.module(package, module, working_dir=working_dir)
    return module


def get_refs(working_dir=None, module=None) -> dict:
    """

    :param working_dir: Default value = None)
    :param module: Default value = None)

    """
    refs = dict(
        classes={
            k: v for k, v in module.classes.items() if not module.classes[k].is_alias
        },
        modules={
            k: v for k, v in module.modules.items() if not module.modules[k].is_alias
        },
        functions=OrderedDict(
            [
                [k, v]
                for k, v in sorted(module.functions.items())
                if not module.functions[k].is_alias
            ]
        ),
    )
    for name, kls in refs["classes"].items():
        annotate.klass(name, kls)
    for name, mod in refs["modules"].items():
        annotate.module(name, mod, working_dir=working_dir)
    for name, fxn in refs["functions"].items():
        annotate.function(name, fxn)
    return refs


def visit_module(
    output=[],
    stats={},
    module=None,
    template=None,
    visited=[],
    exclude: list = [],
    module_name=None,
    working_dir=WORKING_DIR,
):
    """recursive visitor for this package, submodules, classes, functions, etc

    :param output: Default value = [])
    :param stats: Default value = {})
    :param module: Default value = None)
    :param template: Default value = None)
    :param visited: Default value = [])
    :param exclude: list:  (Default value = [])
    :param module_name: Default value = None)
    :param working_dir: Default value = WORKING_DIR)
    :param exclude: list:  (Default value = [])

    """
    if module_name in exclude:
        LOGGER.debug(f"skipping module: {module_name}")
        return output
    annotate.module(module_name, module, working_dir=working_dir)
    refs = get_refs(working_dir=working_dir, module=module)
    # LOGGER.debug(f"exclude: {exclude}")
    LOGGER.debug(f"rendering module: {module_name}")
    rendered = template.render(
        griffe=griffe,
        stats=stats,
        working_dir=working_dir,
        module_name=module_name,
        module=module,
        **refs,
    )
    output.append(clean_text(rendered))
    for name, sub in refs["modules"].items():
        if sub in visited:
            continue
        visit_module(
            output=output,
            module=sub,
            working_dir=working_dir,
            module_name=f"{module_name}.{name}",
            visited=visited + [module],
            exclude=exclude,
            template=template,
        )
    return output


class Checker(mccabe.McCabeChecker):
    """ """

    def run(self):
        if self.max_complexity < 0:
            return
        visitor = mccabe.PathGraphingAstVisitor()
        visitor.preorder(self.tree, visitor)
        for graph in visitor.graphs.values():
            tmp = graph.complexity()
            if tmp > self.max_complexity:
                text = self._error_tmpl % (graph.entity, tmp)
                yield tmp, graph.lineno, graph.column, text, type(self)


def complexity(code: str = None, fname: str = None, threshold: int = 7):
    """

    :param code: str:  (Default value = None)
    :param fname: str:  (Default value = None)
    :param threshold: int:  (Default value = 7)
    :param code: str:  (Default value = None)
    :param fname: str:  (Default value = None)
    :param threshold: int:  (Default value = 7)

    """
    threshold = 7
    try:
        tree = compile(code, fname, "exec", ast.PyCF_ONLY_AST)
    except SyntaxError:
        e = sys.exc_info()[1]
        sys.stderr.write(f"Unable to parse {fname}: {e}\n")
        return 0
    complex = []
    Checker.max_complexity = threshold
    for complexity, lineno, _offset, text, check in Checker(tree, fname).run():
        complex.append(
            dict(
                file=os.path.relpath(fname),
                lineno=lineno,
                # text=text,
                score=complexity,
            )
        )
    out = []
    for admonition in complex:
        out.append(
            dict(
                glyph=GLYPH_COMPLEXITY,
                hover=f'score {admonition["score"]} / {threshold}',
                link=f'/{admonition["file"]}#L{admonition["lineno"]}',
            )
        )
    return out



""" pynchon.util.files
"""

import re
import glob
import functools

import shil

from pynchon import abcs, cli

from pynchon.util import lme, os, typing  # noqa

from .diff import diff, diff_percent, diff_report  # noqa

LOGGER = lme.get_logger(__name__)


@cli.click.argument("prepend_file", nargs=1)
@cli.click.argument("target_file", nargs=1)
@cli.click.option(
    "--clean",
    is_flag=True,
    default=False,
    help="when provided, prepend_file will be removed afterwards",
)
def is_prefix(
    prepend_file: str = None,
    target_file: str = None,
    clean: bool = False,  # noqa
) -> str:
    """
    True if given file already prepends given target
    """
    with open(prepend_file) as fhandle:
        prep_c = fhandle.read()
    with open(target_file) as fhandle:
        target_c = fhandle.read()
    return target_c.lstrip().startswith(prep_c.lstrip())


@cli.click.argument("prepend_file", nargs=1)
@cli.click.argument("target_file", nargs=1)
@cli.click.option(
    "--clean",
    is_flag=True,
    default=False,
    help="when provided, prepend_file will be removed afterwards",
)
def prepend(
    prepend_file: str = None,
    target_file: str = None,
    clean: bool = False,  # noqa
) -> bool:
    """
    Prepends given file contents to given target
    """
    if not is_prefix(prepend_file=prepend_file, target_file=target_file):
        with open(prepend_file) as fhandle:
            prep_c = fhandle.read()
        with open(target_file) as fhandle:
            target_c = fhandle.read()
        with open(target_file, "w") as fhandle:
            fhandle.write(f"""{prep_c}\n{target_c}""".lstrip())
        return True
    return False


def find_suffix(root: str = "", suffix: str = "") -> typing.StringMaybe:
    """ """
    assert root and suffix
    return shil.invoke(f"{root} -type f -name *.{suffix}", strict=True).stdout.split(
        "\n"
    )


@functools.lru_cache(maxsize=None)
def get_git_root(path: str = ".") -> typing.StringMaybe:
    """ """
    path = abcs.Path(path).absolute()
    tmp = path / ".git"
    if tmp.exists():
        return tmp
    elif not path:
        return None
    else:
        try:
            return get_git_root(path.parents[0])
        except IndexError:
            LOGGER.warning(f"Could not find a git-root for '{path}'")


def find_src(
    src_root: str,
    exclude_patterns=[],
    quiet: bool = False,
) -> list:
    """ """
    exclude_patterns = set(list(map(re.compile, exclude_patterns)))
    globs = [
        abcs.Path(src_root).joinpath("**/*"),
    ]
    quiet or LOGGER.info(f"finding src under {globs}")
    globs = [glob.glob(str(x), recursive=True) for x in globs]
    matches = functools.reduce(lambda x, y: x + y, globs)
    matches = [str(x.absolute()) for x in map(abcs.Path, matches) if not x.is_dir()]
    # LOGGER.debug(matches)
    matches = [
        m for m in matches if not any([p.match(str(m)) for p in exclude_patterns])
    ]
    return matches


@typing.validate_arguments
def find_globs(
    globs: typing.List[abcs.Path],
    includes=[],
    logger: object = None,
    quiet: bool = False,
) -> typing.List[str]:
    """ """
    logger = logger or LOGGER
    quiet or logger.info(f"finding files matching {globs}")
    globs = [glob.glob(str(x), recursive=True) for x in globs]
    matches = functools.reduce(lambda x, y: x + y, globs, [])
    for i, m in enumerate(matches):
        for d in includes:
            if abcs.Path(d).has_file(m):
                includes.append(m)
    result = []
    for m in matches:
        assert m
        if m not in includes:
            result.append(abcs.Path(m))
    return result


def dumps(
    content: str = None, file: str = None, quiet: bool = True, logger=LOGGER.info
) -> None:
    """ """
    quiet or logger(f"\n{content}")
    with open(file, "w") as fhandle:
        fhandle.write(content)
    quiet or logger(f'Wrote "{file}"')



""" pynchon.plugins.github
"""

import webbrowser

import shimport
from fleks import cli, tagging

from pynchon import abcs, events, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)
config = shimport.lazy("pynchon.config")
option_api_token = cli.click.option(
    "--api-token", "-t", "token", default="", help="defaults to $GITHUB_API_TOKEN"
)


@tagging.tags(click_aliases=["gh"])
class GitHub(models.ToolPlugin):
    """Tools for working with GitHub"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "github"
        enterprise: bool = typing.Field(default=False)
        org_name: str = typing.Field(default=None)
        org_url: str = typing.Field(default=None)
        repo_url: str = typing.Field(default=None)
        actions: typing.List[abcs.Path] = typing.Field(default=[None])
        raw_url: str = typing.Field(default=None)

        @property
        def raw_url(self):
            return f"https://raw.githubusercontent.com/{self.org_name}/{config.git.repo_name}"

        @property
        def actions(self) -> typing.List[typing.Dict]:
            """ """
            wflows = abcs.Path(config.git.root) / ".github" / "workflows"
            if wflows.exists():
                return [
                    dict(
                        name=fname,
                        file=wflows / fname,
                        url=f"{self.repo_url}/actions/workflows/{fname}",
                    )
                    for fname in wflows.list()
                ]
            else:
                return []

        @property
        def repo_url(self):
            return config.git.repo_url

        @property
        def org_url(self):
            return f"https://github.com/{self.org_name}"

        @property
        def org_name(self):
            return config.git.github_org

    name = "github"
    cli_name = "github"
    cli_aliases = []

    @cli.click.option("--org", "-o")
    def open(self, org=None):
        """Opens org/repo github in a webbrowser

        :param org: Default value = None)

        """
        org_name = self["org_name"]
        url = self["org_url"]
        if not org:
            repo_name = self[:"git.repo_name":]
            url = f"{url}/{repo_name}"
        return webbrowser.open(url)

    @cli.options.org_name
    @option_api_token
    def clone_org(self, org_name: str = None, token: str = None):  # noqa
        """Clones an entire github-org

        :param org_name: str:  (Default value = None)
        :param token: str:  (Default value = None)
        :param org_name: str:  (Default value = None)
        :param token: str:  (Default value = None)

        """
        raise NotImplementedError()

    @cli.click.argument("repo")
    @option_api_token
    def clone(self, repo: str, token: str = None):  # noqa
        """Clones a single repo from this project's org

        :param repo: str:
        :param token: str:  (Default value = None)
        :param repo: str:
        :param token: str:  (Default value = None)

        """
        raise NotImplementedError()

    # @cli.click.argument('repo')
    @tagging.tags(click_aliases=["pr"])
    @option_api_token
    def pull_request(self, repo: str, token: str = None):  # noqa
        """Creates a pull-request from this branch

        :param repo: str:
        :param token: str:  (Default value = None)
        :param repo: str:
        :param token: str:  (Default value = None)

        """
        raise NotImplementedError()

    @tagging.tags(click_aliases=["codeowners"])
    # @option_api_token
    def code_owners(self, repo: str, token: str = None):  # noqa
        """Describes code-owners for changes or for working-dir

        :param repo: str:
        :param token: str:  (Default value = None)
        :param repo: str:
        :param token: str:  (Default value = None)

        """
        raise NotImplementedError()



""" pynchon
"""

# NB: this should have been set by CI immediately
# before pypi-upload.  but the `except` below might
# be triggered by local development.
try:
    from ._version import __version__  # noqa
except ImportError:
    __version__ = "0.0.0+local"



""" setup.py for {{name}}
"""

import os

from setuptools import setup

GIT_COMMIT = os.environ.get("GIT_COMMIT", "local")
if os.environ.get("PYPI_RELEASE", None):
    USE_CALVER = f"%Y.%m.%d.%H.%M"
else:
    USE_CALVER = f"%Y.%m.%d+{GIT_COMMIT}"

if __name__ == "__main__":
    try:
        setup(
            # use_calver=f"%Y.%m.%d.%H.%M+{GIT_COMMIT}",
            use_calver=USE_CALVER,
            setup_requires=["setuptools", "calver"],
        )
    except:  # noqa
        print(
            "\n\nAn error occurred while building the project, "
            "please ensure you have the most updated version of setuptools, "
            "setuptools_scm and wheel with:\n"
            "   pip install -U setuptools setuptools_scm wheel\n\n"
        )
        raise



""" pynchon.models.plugins.pynchon """

import collections

import fleks
import shimport
from fleks import tagging

from pynchon.plugins import util as plugins_util

from . import validators

from pynchon import abcs, api, cli, events  # noqa
from pynchon.util import lme, typing  # noqa


LOGGER = lme.get_logger(__name__)
classproperty = fleks.util.typing.classproperty
pydash = shimport.lazy("pydash")
config_mod = shimport.lazy("pynchon.config")


@tagging.tags(cli_label="<<Abstract>>")
class PynchonPlugin(fleks.Plugin):
    """Pynchon-specific plugin-functionality"""

    name = "<<Abstract>>"
    cli_label = "<<Abstract>>"
    config_class = None
    __class_validators__ = [
        validators.require_conf_key,
        validators.warn_config_kls,
    ]

    @classproperty
    def siblings(kls) -> collections.OrderedDict:
        """
        Returns a dictionary of other plugins for this runtime
        """
        result = []
        from pynchon.plugins import registry

        for plugin in registry.keys():
            if plugin == kls.name:
                continue
            result.append(plugins_util.get_plugin_obj(plugin))
        result = sorted(result, key=lambda p: p.priority)

        class Siblings(collections.OrderedDict):
            def collect_config_list(self, key):
                """collects the named key across all sibling-plugins"""
                out = []
                for name, plugin_obj in self.items():
                    out += plugin_obj[key::[]]
                return out

            def collect_config_dict(self, key):
                """collects the named key across all sibling-plugins"""
                out = {}
                for name, plugin_obj in self.items():
                    out.update(plugin_obj[key::{}])
                return out

        return Siblings([p.name, p] for p in result)

    @classproperty
    def instance(kls):
        """Returns the (singleton) instance for this plugin"""
        return plugins_util.get_plugin_obj(kls.name)

    @classproperty
    def plugin_templates_prefix(kls):
        return f"pynchon/plugins/{kls.config_class.config_key}"

    @classproperty
    def plugin_templates_root(self):
        """ """
        return abcs.Path(self.plugin_templates_prefix)

    @classproperty
    def project_config(self):
        """
        Returns finalized config for the whole project
        """
        return api.project.get_config()

    @classmethod
    def get_config_key(kls):
        """
        Returns config key or (normalized) class-name
        """
        default = kls.name.replace("-", "_")
        config_kls = getattr(kls, "config_class", None)
        return getattr(config_kls, "config_key", default) or default

    @classmethod
    def get_current_config(kls):
        """
        Get the current config for this plugin
        """
        conf_class = getattr(kls, "config_class", None)
        if conf_class is None:
            return {}
        conf_key = kls.get_config_key()
        result = getattr(config_mod, conf_key)
        return result

    @property
    def project_root(self):
        proj_conf = self[:"project.subproject":{}] or self[:"project":]
        return proj_conf.get("root", None) or self[:"git.root":]

    @property
    def plugin_config(self):
        """ """
        return self.get_current_config()

    @property
    def config(self):
        """ """
        return self.cfg()

    def cfg(self):
        """Shows current config for this plugin"""
        kls = self.__class__
        conf_class = getattr(kls, "config_class", None)
        conf_class_name = conf_class.__name__ if conf_class else "(None)"
        LOGGER.debug(f"config class: {conf_class_name}")
        LOGGER.debug("current config:")
        result = kls.get_current_config()
        return result

    def __getitem__(self, key: str):
        """
        shortcut for accessing local plugin-config
        """
        if isinstance(key, (slice,)):

            start, stop, step = key.start, key.stop, key.step
            try:
                if start:
                    result = self[start]
                if stop:
                    result = self % stop
                result = result if result is not None else step
            except (KeyError,) as exc:
                if step is not None:
                    return step
                else:
                    raise
            else:
                return result
        else:
            try:
                return getattr(self.config, key)
            except (AttributeError,) as exc:
                try:
                    return self.config[key]
                except (KeyError,) as exc:
                    fallback = pydash.get(self.config, key)
                    if fallback:
                        return fallback
                    else:
                        raise

    def __floordiv__(self, key: str, strict=False):
        """

        :param key: str:
        :param strict: Default value = False)
        """
        # return self.__mod__(key, strict=strict)
        from pynchon import api

        assert key
        key = key[1:] if key.startswith("/") else key
        return api.render.get_template(f"{self.plugin_templates_prefix}/{key}")

    def __mod__(self, key: str, strict=True):
        """shortcut for accessing global pynchon-config

        :param key: str:
        :param strict: Default value = True)

        """
        try:
            return self.project_config[key]
        except (KeyError,) as exc:
            fallback = pydash.get(self.project_config, key, None)
            if fallback is not None:
                return fallback
            else:
                if strict:
                    raise



""" pynchon.util.text.dumpf

    Helpers for loading data structures from files
"""

import sys

from pynchon import cli

from pynchon.util import lme, text, typing  # noqa

LOGGER = lme.get_logger(__name__)


@cli.click.argument("file", default="/dev/stdin")
def json(obj, file=None, **kwargs) -> None:
    """ """
    with open(file, "w") as fhandle:
        fhandle.write(text.dumps.json(obj, **kwargs))


@cli.click.argument("file", default="/dev/stdin")
def yaml(file=None, output=None, **kwargs):
    """
    Parse JSON input and writes YAML
    """
    file = "/dev/stdin" if file == "-" else file
    content = text.dumps.yaml(file=file, **kwargs)
    fhandle = open(output, "w") if output else sys.stdout
    content = f"{content}\n"
    fhandle.write(content)
    fhandle.close()



""" pynchon.plugins.json
"""

from pynchon import abcs, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class Json(models.ToolPlugin):
    """Tools for working with JSON & JSON5"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "json"

    # cli_subsumes: typing.List[typing.Callable] = [
    #     loadf_main.json5,
    #     loadf_main.json,
    #     loadf_main.j5,
    #     # loadf_main.json,
    # ]

    # @tags(click_aliases=['loads',])
    # def json_loads(self):
    #     """ loads JSON from string-input (strict) """

    # @tags(click_aliases=['loadf',])
    # def json_loadf(self):
    #     """ loads JSON from file-input (strict) """
    #     pass

    # def load_json5(self):
    #     """ loads JSON-5 from string-input """
    #     pass
    #
    # def loadf_json5(self):
    #     """ loads JSON-5 from file-input """
    #     pass
    # DEFAULT_OPENER = "open"
    #
    #
    # @kommand(
    #     name="any",
    #     parent=PARENT,
    #     formatters=dict(
    #         # markdown=pynchon.T_TOC_CLI,
    #     ),
    #     options=[
    #         # options.file,
    #         options.format,
    #         # options.stdout,
    #         options.output,
    #     ],
    # )
    # def render_any(format, file, stdout, output):
    #     """
    #     Render files with given renderer
    #     """
    #     raise NotImplementedError()

    name = "json"
    priority = 1
    cli_name = name



""" pynchon.api.python
"""

import functools

from pynchon.util import files, lme, text

LOGGER = lme.get_logger(__name__)


@functools.lru_cache(maxsize=None)
def is_package(folder: str) -> bool:
    """slightly better than just looking for setup.py-
    we try to use it to get the current version-string

    :param folder: str:
    """
    from pynchon.util.os import invoke

    cmd = invoke(
        f"cd {folder} && python setup.py --version 2>/dev/null", log_command=False
    )
    return cmd.succeeded


def load_setupcfg(file: str = "", folder: str = ""):
    """

    :param file: str:  (Default value = "")
    :param folder: str:  (Default value = "")
    :param file: str:  (Default value = "")
    :param folder: str:  (Default value = "")

    """
    if not file:
        folder = folder or files.get_git_root().parents[0]
        file = folder / "setup.cfg"
    return text.loadf.ini(file)


from pynchon.util import typing


def load_entrypoints(config: dict = None) -> typing.List[typing.Dict]:
    """
    loads entrypoints from setup.cfg style configuration
    """
    if not config:
        LOGGER.critical("no config provided!")
        return {}
    try:
        console_scripts = config["options.entry_points"]["console_scripts"]
    except (KeyError,) as exc:
        LOGGER.warning(
            f'could not load config["options.entry_points"]["console_scripts"] from {config}'
        )
        return {}
    console_scripts = [x for x in console_scripts.split("\n") if x]
    package = config["metadata"]["name"]
    entrypoints = []
    for c in console_scripts:
        tmp = dict(
            package=package,
            bin_name=c.split("=")[0].strip(),
            module=c.split("=")[1].strip().split(":")[0],
            entrypoint=c.split("=")[1].strip().split(":")[1],
        )
        abs_entrypoint = tmp["module"] + ":" + tmp["entrypoint"]
        tmp["setuptools_entrypoint"] = abs_entrypoint
        entrypoints.append(tmp)
    return entrypoints
    # return dict(package=package,entrypoints=entrypoints,)



""" pynchon.plugins.release
"""

# from pynchon import models
# from pynchon.util import lme, typing
#
# LOGGER = lme.get_logger(__name__)
#
#
# class Release(models.Tool):
#     """
#     Namespace for rendering docs-generation commands from other plugins
#     """
#
#     name = cli_name = 'gen'
#     priority = -1
#     config_class = None
#     cli_subsumes: typing.List[typing.Callable] = []
#
#     # @common.groop("api", parent=groups.gen)
#     # def gen_api() -> None:
#     #     """
#     #     Generate API docs from python modules, packages, etc
#     #     """
#     #
#     # @common.groop("cli", parent=groups.gen)
#     # def gen_cli():
#     #     """Generate CLI docs"""
#     #
